import base64
import glob
import json
import os
import shutil
import tempfile
from unittest import mock

import yaml

from tests.unit import base
from chirp import chirp_common
from chirp import directory


class TestDirectory(base.BaseTest):
    def cleanUp(self):
        shutil.rmtree(self.tempdir)

    def setUp(self):
        super(TestDirectory, self).setUp()
        self.tempdir = tempfile.mkdtemp()
        directory.enable_reregistrations()

        class FakeAlias(chirp_common.Alias):
            VENDOR = 'Taylor'
            MODEL = 'Barmaster 2000'
            VARIANT = 'A'

        @directory.register
        class FakeRadio(chirp_common.CloneModeRadio):
            VENDOR = 'Dan'
            MODEL = 'Foomaster 9000'
            VARIANT = 'R'
            ALIASES = [FakeAlias]

            @classmethod
            def match_model(cls, file_data, image_file):
                return file_data == b'thisisrawdata'

        self.test_class = FakeRadio

    def _test_detect_finds_our_class(self, tempfn):
        radio = directory.get_radio_by_image(tempfn)
        self.assertTrue(isinstance(radio, self.test_class))
        return radio

    def test_detect_with_no_metadata(self):
        fn = os.path.join(self.tempdir, 'testfile')
        with open(fn, 'wb') as f:
            f.write(b'thisisrawdata')
            f.flush()
        self._test_detect_finds_our_class(fn)

    def test_detect_with_metadata_base_class(self):
        fn = os.path.join(self.tempdir, 'testfile')
        with open(fn, 'wb') as f:
            f.write(b'thisisrawdata')
            f.write(self.test_class.MAGIC + b'-')
            f.write(self.test_class(None)._make_metadata())
            f.flush()
        self._test_detect_finds_our_class(fn)

    def test_detect_with_metadata_alias_class(self):
        fn = os.path.join(self.tempdir, 'testfile')
        with open(fn, 'wb') as f:
            f.write(b'thisisrawdata')
            f.write(self.test_class.MAGIC + b'-')
            FakeAlias = self.test_class.ALIASES[0]
            fake_metadata = base64.b64encode(json.dumps(
                {'vendor': FakeAlias.VENDOR,
                 'model': FakeAlias.MODEL,
                 'variant': FakeAlias.VARIANT,
                 }).encode())
            f.write(fake_metadata)
            f.flush()
        radio = self._test_detect_finds_our_class(fn)
        self.assertEqual('Taylor', radio.VENDOR)
        self.assertEqual('Barmaster 2000', radio.MODEL)
        self.assertEqual('A', radio.VARIANT)


class TestDetectBruteForce(base.BaseTest):
    def test_detect_all(self):
        # Attempt a brute-force detection of all test images.
        #
        # This confirms that no test image is detected by more than one
        # radio class. If it is, fail and report those classes.

        path = os.path.dirname(__file__)
        path = os.path.join(path, '..', 'images', '*.img')
        test_images = glob.glob(path)
        self.assertNotEqual(0, len(test_images))
        for image in test_images:
            detections = []
            filedata = open(image, 'rb').read()
            for cls in directory.RADIO_TO_DRV:
                if not hasattr(cls, 'match_model'):
                    continue
                if cls.match_model(filedata, image):
                    detections.append(cls)
            if len(detections) > 1:
                raise Exception('Detection of %s failed: %s' % (image,
                                                                detections))


class TestAliasMap(base.BaseTest):
    def test_uniqueness(self):
        directory_models = {}
        for rclass in directory.DRV_TO_RADIO.values():
            for cls in [rclass] + rclass.ALIASES:
                # Make sure there are no duplicates
                directory_models.setdefault(cls.VENDOR, set())
                fullmodel = '%s%s' % (cls.MODEL, cls.VARIANT)
                self.assertNotIn(fullmodel,
                                 directory_models[cls.VENDOR])
                directory_models[cls.VENDOR].add(fullmodel)

        aliases = yaml.load(open(os.path.join(os.path.dirname(__file__),
                                              '..', '..', 'chirp', 'share',
                                              'model_alias_map.yaml')).read(),
                            Loader=yaml.FullLoader)
        for vendor, models in sorted(aliases.items()):
            directory_models.setdefault(vendor, set())
            my_aliases = set([x['model'] for x in models])
            vendor = vendor.split('/')[0]
            for model in models:
                # Make sure the thing we tell users to use is in the
                # directory
                try:
                    alt_vendor, alt_model = model['alt'].split(' ', 1)
                except ValueError:
                    alt_vendor = vendor
                    alt_model = model['alt']

                # Aliases may reference other aliases?
                self.assertIn(alt_model,
                              directory_models[alt_vendor] | my_aliases,
                              '%s %s not found for %s %s' % (
                                  alt_vendor, alt_model,
                                  vendor, model['model']))

                # Make sure the alias model is NOT in the directory
                # before we add it to ensure there are no duplicates
                self.assertNotIn(model['model'], directory_models[vendor])
                directory_models[vendor].add(model['model'])


class TestDetectedBy(base.BaseTest):
    @mock.patch('chirp.directory.DRV_TO_RADIO', new={})
    @mock.patch('chirp.directory.RADIO_TO_DRV', new={})
    def test_detected_isolation(self):
        @directory.register
        class BaseRadio(chirp_common.CloneModeRadio):
            VENDOR = 'CHIRP'
            MODEL = 'Base'

        @directory.register
        class SubRadio1(BaseRadio):
            MODEL = 'Sub1'

        @directory.register
        @directory.detected_by(SubRadio1)
        class SubRadio2(SubRadio1):
            MODEL = 'Sub2'

        # BaseRadio should not think it detects the subs
        self.assertEqual([BaseRadio], BaseRadio.detected_models())

        # Sub1 detects both itself and Sub2
        self.assertEqual([SubRadio1, SubRadio2], SubRadio1.detected_models())

        # If include_self=False, Sub1 should not include itself
        self.assertEqual([SubRadio2],
                         SubRadio1.detected_models(include_self=False))

        # Sub2 does not also detect Sub1
        self.assertEqual([SubRadio2], SubRadio2.detected_models())

    @mock.patch('chirp.directory.DRV_TO_RADIO', new={})
    @mock.patch('chirp.directory.RADIO_TO_DRV', new={})
    def test_detected_include_self(self):
        class BaseRadio(chirp_common.CloneModeRadio):
            VENDOR = 'CHIRP'
            MODEL = 'Base'

        @directory.register
        @directory.detected_by(BaseRadio)
        class SubRadio(BaseRadio):
            MODEL = 'Sub'

        # BaseRadio should not include itself since it is not registered
        self.assertEqual([SubRadio], BaseRadio.detected_models())



import os
import sys
import time
from unittest import mock

import ddt

sys.modules['wx'] = wx = mock.MagicMock()
sys.modules['wx.lib'] = mock.MagicMock()
sys.modules['wx.lib.scrolledpanel'] = mock.MagicMock()
sys.modules['wx.lib.sized_controls'] = mock.MagicMock()
sys.modules['wx.richtext'] = mock.MagicMock()
wx.lib.newevent.NewCommandEvent.return_value = None, None
sys.modules['chirp.wxui.developer'] = mock.MagicMock()

# These need to be imported after the above mock so that we don't require
# wx to be present for these tests
from tests.unit import base  # noqa
from chirp import chirp_common  # noqa
from chirp import directory  # noqa
from chirp.wxui import clone  # noqa
from chirp.wxui import config  # noqa
from chirp.wxui import radiothread  # noqa


class TestRadioThread(base.BaseTest):
    def setUp(self):
        super().setUp()

    def test_radiojob(self):
        radio = mock.MagicMock()
        editor = mock.MagicMock()
        job = radiothread.RadioJob(editor, 'get_memory', [12], {})
        self.assertIsNone(job.dispatch(radio))
        radio.get_memory.assert_called_once_with(12)
        self.assertEqual(job.result, radio.get_memory.return_value)

    def test_radiojob_exception(self):
        radio = mock.MagicMock()
        radio.get_memory.side_effect = ValueError('some error')
        editor = mock.MagicMock()
        job = radiothread.RadioJob(editor, 'get_memory', [12], {})
        self.assertIsNone(job.dispatch(radio))
        radio.get_memory.assert_called_once_with(12)
        self.assertIsInstance(job.result, ValueError)

    def test_thread(self):
        radio = mock.MagicMock()
        radio.get_features.side_effect = ValueError('some error')
        editor = mock.MagicMock()
        # Simulate an edit conflict with the first event by returning
        # False for "delivered" to force us to queue an event.
        editor.radio_thread_event.side_effect = [False, True, True, True]
        thread = radiothread.RadioThread(radio)
        mem = mock.MagicMock()
        job1id = thread.submit(editor, 'get_memory', 12)
        job2id = thread.submit(editor, 'set_memory', mem)
        job3id = thread.submit(editor, 'get_features')
        # We have to start the thread after we submit the main jobs so
        # the order is stable for comparison.
        thread.start()

        # Wait for the main jobs to be processed before we signal exit
        while not all([radio.get_memory.called,
                       radio.set_memory.called,
                       radio.get_features.called]):
            time.sleep(0.1)

        thread.end()
        thread.join(5)
        self.assertFalse(thread.is_alive())
        radio.get_memory.assert_called_once_with(12)
        radio.set_memory.assert_called_once_with(mem)
        radio.get_features.assert_called_once_with()
        self.assertEqual(4, editor.radio_thread_event.call_count)

        # We expect the jobs to be delivered in order of
        # priority. Since we return False for the first call to
        # radio_thread_event(), job2 should be queued and then
        # delivered first on the next cycle.
        expected_order = [job2id, job2id, job3id, job1id]
        for i, (jobid, call) in enumerate(
                zip(expected_order,
                    editor.radio_thread_event.call_args_list)):
            job = call[0][0]
            self.assertEqual(jobid, job.id)

        # We should call non-blocking for every call except the last
        # one, when the queue is empty
        editor.radio_thread_event.assert_has_calls([
            mock.call(mock.ANY, block=False),
            mock.call(mock.ANY, block=False),
            mock.call(mock.ANY, block=False),
            mock.call(mock.ANY, block=True),
        ])

    def test_thread_abort_priority(self):
        radio = mock.MagicMock()
        radio.get_features.side_effect = ValueError('some error')
        editor = mock.MagicMock()
        thread = radiothread.RadioThread(radio)
        mem = mock.MagicMock()
        thread.submit(editor, 'get_memory', 12)
        thread.submit(editor, 'set_memory', mem)
        thread.submit(editor, 'get_features')
        thread.end()
        # We have to start the thread after we submit the main jobs so
        # the order is stable for comparison.
        thread.start()

        thread.join(5)
        self.assertFalse(thread.is_alive())

        # Our end sentinel should have gone to the head of the queue
        # so that exiting the application does not leave a thread
        # running in the background fetching hundreds of memories.
        radio.get_memory.assert_not_called()
        radio.set_memory.assert_not_called()
        radio.get_features.assert_not_called()
        wx.PostEvent.assert_not_called()


class TestClone(base.BaseTest):
    @mock.patch('platform.system', return_value='Linux')
    def test_sort_ports_unix(self, system):
        ports = [
            mock.MagicMock(device='/dev/cu.zed',
                           description='My Zed'),
            mock.MagicMock(device='/dev/cu.abc',
                           description='Some device'),
            mock.MagicMock(device='/dev/cu.serial',
                           description='')
            ]
        self.assertEqual(
            ['Some device (cu.abc)',
             'cu.serial',
             'My Zed (cu.zed)'],
            [clone.port_label(p)
                for p in sorted(ports, key=clone.port_sort_key)])

    @mock.patch('platform.system', return_value='Windows')
    def test_sort_ports_windows(self, system):
        ports = [
            mock.MagicMock(device='COM7',
                           description='Some serial device'),
            mock.MagicMock(device='COM17',
                           description='Some other device'),
            mock.MagicMock(device='CNC0',
                           description='Some weird device'),
            mock.MagicMock(device='COM4',
                           description=''),
            ]
        self.assertEqual(
            ['CNC0: Some weird device',
             'COM4',
             'COM7: Some serial device',
             'COM17: Some other device'],
            [clone.port_label(p)
                for p in sorted(ports, key=clone.port_sort_key)])

    def test_detected_model_labels(self):
        # Make sure all our detected model labels will be reasonable
        # (and nonzero) in length. If the full label is too long, it will not
        # be visible in the model box.
        for rclass in [x for x in directory.DRV_TO_RADIO.values()
                       if issubclass(x, chirp_common.DetectableInterface)]:
            label = clone.get_model_label(rclass)
            self.assertLessEqual(len(label), 32,
                                 'Label %r is too long' % label)


class TestException(Exception):
    pass


@ddt.ddt
class TestStartup(base.BaseTest):
    def setUp(self):
        super().setUp()
        self.use(mock.patch('os.path'))
        self.use(mock.patch('os.makedirs'))
        self.use(mock.patch('chirp.wxui.CONF'))
        self.args = mock.MagicMock()
        self.args.install_desktop_app = False
        self.args.no_install_desktop_app = False
        from chirp.wxui import maybe_install_desktop, CONF
        self.maybe_install_desktop = maybe_install_desktop
        self.conf = CONF

    @ddt.data(
        # No arguments, no file, no previous, answer no
        [False, False, False, False, False],
        # No arguments, no file, no previous, answer yes
        [False, False, False, False, True],
        # No arguments, no file, previous yes, no prompt
        [False, False, False, True, None],
        # No arguments, exists, previous no, no prompt
        [False, False, True, False, None],
        # Opt out, no file, no prompt
        [False, True, False, None, None],
        # Opt in, no file, previous yes, still prompt
        [True, False, False, True, True],
        # Opt in, exists, previous no, no prompt'),
        [True, False, True, False, None],
    )
    @ddt.unpack
    def test_linux_desktop_file(self, optin, optout, exists, last, answ):
        self.args.install_desktop_app = optin
        self.args.no_install_desktop_app = optout
        os.path.exists.return_value = exists
        self.conf.get_bool.return_value = last
        wx.MessageBox.return_value = wx.YES if answ else wx.NO
        os.makedirs.side_effect = TestException
        wx.MessageBox.reset_mock()

        if answ is True:
            # If we made it through all the checks, and thus prompted the user,
            # make sure we get to the makedirs part if expected
            self.assertRaises(TestException,
                              self.maybe_install_desktop, self.args, None)
        elif answ is False:
            # If we were supposed to make it to the prompt but answer no,
            # make sure we did
            self.maybe_install_desktop(self.args, None)
            self.assertFalse(os.makedirs.called)
            self.assertTrue(wx.MessageBox.called)
        else:
            # If we were not supposed to make it to the prompt, make sure we
            # didn't, nor did we do any create actions
            self.maybe_install_desktop(self.args, None)
            self.assertFalse(os.makedirs.called)
            self.assertFalse(wx.MessageBox.called)

    def test_linux_desktop_file_exists(self):
        os.path.exists.return_value = True
        self.maybe_install_desktop(self.args, None)
        self.assertFalse(os.makedirs.called)



# Copyright 2024 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import logging
import os
import re

import lark
import wx
import wx.adv
from wx import GetTranslation as _

from chirp import chirp_common

LOG = logging.getLogger(__name__)
LANG = """
start: qexpr
qexpr: qexpr (OPERATOR qexpr)* | "(" qexpr ")" | _expr
QUOTE: /"/
PROPERTY: /[a-z0-9_]+/
TEXT: /[^"]+/
INT: /[0-9]+/
FLOAT: /[0-9]+\\.[0-9]+/
OPERATOR: "AND"i | "OR"i
value: QUOTE TEXT QUOTE | INT | FLOAT
_list: "[" value ("," value)* "]"
_expr: equal | match | contains | range
equal: PROPERTY "=" value
contains: PROPERTY "IN"i _list
match: PROPERTY "~" value
range: PROPERTY "<" value "," value ">"
%ignore " "
"""

MEM_FIELDS_SKIP = ['CSV_FORMAT', 'empty', 'extd_number', 'immutable', 'vfo',
                   'extra']
MEM_FIELDS = [x for x in dir(chirp_common.Memory)
              if not x.startswith('_') and x not in MEM_FIELDS_SKIP and
              not inspect.isfunction(getattr(chirp_common.Memory, x))]


def union(a, b):
    return list({x.number: x for x in a + b}.values())


def intersection(a, b):
    only = {x.number for x in a} & {x.number for x in b}
    return list({x.number: x for x in a + b if x.number in only}.values())


class Interpreter(lark.Transformer):
    def __init__(self, memories, visit_tokens: bool = True) -> None:
        self._memories = memories
        super().__init__(visit_tokens)

    def _val(self, tree):
        if isinstance(tree, lark.Token):
            value = tree
        else:
            value = tree.children[0]
        if value.type == 'TEXT':
            return value.value
        elif value.type == 'INT':
            return int(value.value)
        elif value.type == 'FLOAT':
            return float(value.value)
        else:
            raise RuntimeError('Unsupported value type %r' % value.name)

    def _get(self, mem, key):
        if key == 'freq':
            return mem.freq / 1000000
        return getattr(mem, key)

    def equal(self, items):
        prop = items[0].value
        value = self._val(items[1])
        return [x for x in self._memories if self._get(x, prop) == value]

    def contains(self, items):
        prop = items[0].value
        opts = [self._val(x) for x in items[1:]]
        return [x for x in self._memories if self._get(x, prop) in opts]

    def match(self, items):
        prop = items[0].value
        value = self._val(items[1])
        return [x for x in self._memories
                if re.search(value, self._get(x, prop), re.IGNORECASE)]

    def range(self, items):
        prop = items[0].value
        lo = self._val(items[1])
        hi = self._val(items[2])
        return [x for x in self._memories if lo <= self._get(x, prop) <= hi]

    def value(self, items):
        if len(items) == 1:
            # value
            return items[0]
        else:
            # " value "
            return items[1]

    def OPERATOR(self, item):
        if item == 'OR':
            return union
        else:
            return intersection

    def qexpr(self, items):
        while len(items) > 2:
            left = items.pop(0)
            op = items.pop(0)
            right = items.pop(0)
            result = op(left, right)
            items.insert(0, result)
        return items[0]


class QueryFilterError(SyntaxError):
    pass


class PropertyNameError(QueryFilterError):
    label = _('Property Name')


class OperatorError(QueryFilterError):
    label = _('Operator')


class PropertyValueError(QueryFilterError):
    label = _('Property Value')


class PropertyValueStringError(QueryFilterError):
    label = _('Close String')


class PropertyValueFloatError(QueryFilterError):
    label = _('Finish Float')


class FilteringError(Exception):
    pass


class SearchHelp(wx.PopupTransientWindow):
    def __init__(self, parent):
        super().__init__(parent, flags=wx.SIMPLE_BORDER)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self._msg = wx.StaticText(self)
        vbox.Add(self._msg, 1, flag=wx.EXPAND | wx.ALL, border=10)
        link = wx.adv.HyperlinkCtrl(
            self, label=_("Query syntax help"),
            url='https://chirpmyradio.com/projects/chirp/wiki/QueryStrings')
        vbox.Add(link, 1, flag=wx.EXPAND | wx.ALL, border=10)
        bgc = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE)
        self.SetBackgroundColour(bgc)

    def Show(self, show=True):
        parent = self.GetParent()
        psize = parent.GetSize()
        self.Position(parent.ClientToScreen((0, 0)), (10, psize[1]))
        self.SetMinSize((psize[0] - 40, 100))
        self.Fit()
        return super().Show(show)

    def clear(self):
        if self.GetParent().valid:
            self._msg.SetLabel(_('Query syntax OK'))
        else:
            self._msg.SetLabel(_('Type a simple search string or a formatted '
                                 'query and press enter'))
        self.Show(self.IsShown())

    def eat_error(self, exc, orig):
        try:
            expected = set(orig.expected)
        except AttributeError:
            try:
                expected = set(orig.allowed)
            except AttributeError:
                expected = []

        label = []
        if isinstance(exc, PropertyNameError):
            label.append(_('Memory field name (one of %s)') % ','.join(
                sorted(MEM_FIELDS)))
        elif isinstance(exc, PropertyValueError):
            label.append(_('Value to search memory field for'))
        elif isinstance(exc,
                        PropertyValueStringError) or expected == ['QUOTE']:
            label.append(_('Close string value with double-quote (")'))
        elif isinstance(exc, PropertyValueFloatError):
            label.append(_('Finish float value like: 146.52'))
        elif not isinstance(exc, (lark.exceptions.ParseError,
                                  lark.exceptions.UnexpectedCharacters)):
            label.append(str(exc))

        LOG.debug('Query string %r allowed next: %r',
                  self.GetParent().GetValue(), expected)

        exp_map = {
            'QUOTE': '"',
            'LESSTHAN': '<',
            'MORETHAN': '>',
            'LPAR': '(',
            'RPAR': ')',
            'TILDE': '~',
            'EQUAL': '=',
            'LSQB': '[',
            'RSQB': ']',
        }
        expected_filtered = sorted([exp_map.get(e, e) for e in expected])
        if expected:
            label.append('Expected: %s' % (','.join(expected_filtered)))

        for exp in sorted(expected):
            if exp == 'QUOTE':
                label.append(_('Example: "foo"'))
            elif exp == 'INT':
                label.append(_('Example: 123'))
            elif exp == 'FLOAT':
                label.append(_('Example: 146.52'))
            elif exp == 'OPERATOR':
                label.append(_('Example: AND, OR'))
            elif exp in ('LPAR', 'RPAR'):
                label.append(_('Example: ( expression )'))
            elif exp == 'TILDE':
                label.append(_('Example: name~"myrepea"'))
            elif exp == 'EQUAL':
                label.append(_('Example: name="myrepeater'))
            elif exp in ('LESSTHAN', 'MORETHAN'):
                label.append(_('Example: freq<146.0,148.0>'))
            elif exp == 'PROPERTY':
                label.append(_('One of: %s') % ','.join(MEM_FIELDS))
            else:
                LOG.debug(_('No example for %s') % exp)

        self._msg.SetLabel(os.linesep.join(label))
        self.Show()


class SearchBox(wx.TextCtrl):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.Bind(wx.EVT_TEXT, self._typing)
        self.parser = lark.Lark(LANG)
        self.query = None
        self.help = SearchHelp(self)
        self.help.Hide()
        self.help.clear()
        self.Bind(wx.EVT_SIZE, self._resize)
        self.SetHint(_('Filter...'))
        self.Bind(wx.EVT_SET_FOCUS, self._focus)
        self.Bind(wx.EVT_KILL_FOCUS, self._unfocus)

    def _focus(self, event):
        self.help.Show()
        event.Skip()

    def _unfocus(self, event):
        # Without this delay, if the user clicks on the help link it won't
        # activate because we hide it too quickly (apparently)
        wx.CallLater(100, self.help.Hide)
        event.Skip()

    def _resize(self, event):
        self.help.Show(self.help.IsShown())
        event.Skip()

    _errors = {
        PropertyNameError: ['foo'],
        PropertyValueError: ['foo=', 'foo<', 'foo IN', 'foo~'],
        PropertyValueStringError: ['foo="', 'foo~"'],
        PropertyValueFloatError: ['foo=1.'],
        }

    def _typing(self, event):
        query_string = self.GetValue()
        self.help.Show()
        if not query_string.strip():
            self.help.clear()
            return
        if query_string.isalnum():
            # Assume simple search
            self.help.clear()
            return
        try:
            self.query = self.parser.parse(query_string)
            LOG.debug('Query:\n%s', self.query.pretty())
        except lark.UnexpectedInput as e:
            exc_class = e.match_examples(self.parser.parse, self._errors,
                                         use_accepts=True)
            if not exc_class:
                # No specific match
                exc = e
            else:
                exc = exc_class(e.get_context(query_string), e.line, e.column)
            self.help.eat_error(exc, e)
            self.query = None
        except Exception as e:
            LOG.debug('Parse error: %s (%s)', e, e.__class__)
            self.query = None
        else:
            self.help.clear()

    @property
    def valid(self):
        return self.query is not None

    def filter_memories(self, memories):
        if not self.query:
            raise QueryFilterError(_('Query string is invalid'))
        try:
            r = Interpreter(memories).transform(self.query)
        except lark.exceptions.VisitError as e:
            self.help.eat_error(e.orig_exc, e)
            self.help.Show()
            raise FilteringError(_('Error applying filter'))
        else:
            return r.children[0]



import logging
import threading

from chirp import chirp_common, directory, errors

from chirp.drivers.kenwood_live import KenwoodLiveRadio, NOCACHE, RADIO_IDS

LOG = logging.getLogger(__name__)

TS790E_DUPLEX = ["", "split", "+", "-"]
TS790E_TMODES = ["", "Tone"]
TS790E_SKIP = ["", "S"]

TS790E_MODES = {
    "LSB": "1",
    "USB": "2",
    "CW":  "3",
    "FM":  "4",
    "CWN": "7",
}
TS790E_MODES_REV = {val: mode for mode, val in TS790E_MODES.items()}

TS790E_TONES = list(chirp_common.OLD_TONES)
TS790E_TONES.remove(69.3)
TS790E_TONES.remove(97.4)
TS790E_OFFSETS = ["", "+", "-"]

TS790E_BANDS = [
    (135000000, 173999900),
    (300000000, 949999900),
    (1200000000, 1304957000),
]

TS790E_SHIFTS = [600000, 1600000, 35000000]

TS790E_SPECIALS = ["P1", "P2", "P3", "A1", "A2", "A3", "C1", "C2", "C3"]


def get_band(freq):
    i = 0
    for band in TS790E_BANDS:
        if freq >= band[0] and freq <= band[1]:
            return i
        i += 1
    return None


@directory.register
class TS790ERadio(KenwoodLiveRadio):
    """Kenwood TS-790E"""
    MODEL = "TS-790E"
    BAUD_RATE = 4800
    HARDWARE_FLOW = False
    COMMAND_DELIMITER = (";", "")

    _upper = 49
    _kenwood_valid_tones = list(TS790E_TONES)

    def command_no_output(self, ser, cmd, *args):
        """Send @cmd to radio via @ser"""
        with self.LOCK:
            cmd += "".join(args) + ";"

            LOG.debug("PC->RADIO: %s" % cmd.strip())
            ser.write(cmd.encode('cp1252'))

    def __init__(self, *args, **kwargs):
        chirp_common.LiveRadio.__init__(self, *args, **kwargs)
        self._memcache = {}
        self.LOCK = threading.Lock()
        self.LAST_BAUD = self.BAUD_RATE
        self.LAST_DELIMITER = self.COMMAND_DELIMITER

        if self.pipe:
            self.pipe.timeout = 0.1
            self.pipe.baudrate = 4800
            try:
                resp = self.command(self.pipe, "ID")
            except UnicodeDecodeError:
                raise Exception("Radio is likely not %s" % (self.MODEL))
            radio_id = resp.split(";")[0]
            if RADIO_IDS[radio_id] != \
                    self.MODEL.split(" ")[0]:
                raise Exception("Radio reports %s (not %s)" % (radio_id,
                                                               self.MODEL))
            self.command_no_output(self.pipe, "AI", "0")

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.can_odd_split = True

        rf.has_bank = False
        rf.has_ctone = False
        rf.has_dtcs = False
        rf.has_dtcs_polarity = False
        rf.has_name = False
        rf.has_tuning_step = False
        rf.has_nostep_tuning = True
        rf.memory_bounds = (0, self._upper)

        rf.valid_bands = TS790E_BANDS
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC
        rf.valid_duplexes = TS790E_DUPLEX
        rf.valid_modes = list(TS790E_MODES.keys())
        rf.valid_skips = TS790E_SKIP
        rf.valid_tmodes = TS790E_TMODES
        rf.valid_special_chans = list(TS790E_SPECIALS)

        return rf

    def get_memory(self, number):
        if isinstance(number, str):
            number = 50 + TS790E_SPECIALS.index(number)
        if number < 0 or number > (self._upper + len(TS790E_SPECIALS)):
            raise errors.InvalidMemoryLocation(
                "Number must be between 0 and %i" % (
                    self._upper + len(TS790E_SPECIALS)))
        if number in self._memcache and not NOCACHE:
            return self._memcache[number]

        result = self.command(
            self.pipe, *self._cmd_get_memory(number))
        if result.endswith("00000000000000000") or result.endswith("?"):
            mem = chirp_common.Memory()
            if number >= 30:
                mem.duplex = TS790E_DUPLEX[1]
            if number >= 50:
                mem.extd_number = TS790E_SPECIALS[number - 50]
            mem.number = number
            mem.offset = None
            mem.empty = True
            mem.skip = TS790E_SKIP[0]
            self._memcache[number] = mem
            return mem

        mem = self._parse_mem_spec(result)
        if number >= 30:  # check for split frequency operation
            result = self.command(
                self.pipe, *self._cmd_get_split(number))
            self._parse_split_spec(mem, result)
        self._memcache[number] = mem

        return mem

    def set_memory(self, memory):
        LOG.debug("set_memory(%s)" % repr(memory))
        if memory.extd_number:
            real_number = 50 + TS790E_SPECIALS.index(memory.extd_number)
        else:
            real_number = memory.number
        if real_number < 0 or real_number > (self._upper +
                                             len(TS790E_SPECIALS)):
            raise errors.InvalidMemoryLocation(
                "Number must be between 0 and %i" % (
                    self._upper + len(TS790E_SPECIALS)))

        if memory.extd_number and (memory.extd_number[0] in ["A", "C", "P"]):
            band = get_band(memory.freq)
            supposed_band = int(memory.extd_number[1]) - 1
            if band != supposed_band:
                raise errors.InvalidMemoryLocation(
                        "%s frequency must be between %f and %f " %
                        (memory.extd_number, *TS790E_BANDS[supposed_band]))

        del self._memcache[real_number]

        # If we have a split set the transmit frequency first.
        if real_number >= 30 and memory.offset is not None:
            spec = self._make_base_spec(memory, memory.offset)
            self.command_no_output(
                self.pipe, *self._cmd_set_split(real_number, spec))

        spec = self._make_base_spec(memory, memory.freq)
        self.command_no_output(self.pipe,
                               *self._cmd_set_memory(real_number, spec))

    def erase_memory(self, number):
        if number not in self._memcache and not NOCACHE:
            return
        self.command_no_output(self.pipe, *self._cmd_erase_memory(number))
        del self._memcache[number]

    def _cmd_get_memory(self, number):
        return "MR0 %02i" % number

    def _cmd_get_split(self, number):
        return "MR1 %02i" % number

    def _cmd_set_memory(self, number, spec):
        return "MW0 %02i%s" % (number, spec)

    def _cmd_set_split(self, number, spec):
        return "MW1 %02i%s" % (number, spec)

    def _cmd_erase_memory(self, number):
        return "MW0 %02i%017i" % (number, 0)

    def _parse_mem_spec(self, spec):
        mem = chirp_common.Memory()

        # pad string so indexes match Kenwood docs
        spec = " " + spec   # Param Format Function

        _p3 = spec[5:7]     # P3    7      Memory Channel
        _p4 = spec[7:18]    # P4    4      Frequency
        _p5 = spec[18]      # P5    2      Mode
        _p6 = spec[19]      # P6    10     Memory Lockout
        if len(spec) == 24:
            _p7 = spec[20]      # P7    1      Tone On/Off
            _p8 = spec[21:23]   # P8    14     Tone Frequency
            _p9 = spec[23]      # P9    13     Offset

        if not TS790E_MODES_REV[_p5]:
            return mem
        mem.number = int(_p3)
        mem.duplex = TS790E_DUPLEX[0]
        mem.freq = int(_p4)
        mem.mode = TS790E_MODES_REV[_p5]
        if mem.number < 30 and mem.mode == "FM":
            mem.duplex = TS790E_OFFSETS[int(_p9)]
            if mem.duplex != TS790E_OFFSETS[0]:
                cur_band = get_band(mem.freq)
                mem.offset = TS790E_SHIFTS[cur_band]
            else:
                mem.offset = None

        mem.skip = TS790E_SKIP[int(_p6)]
        if mem.mode == "FM":
            mem.tmode = TS790E_TMODES[int(_p7)]
            if mem.tmode == TS790E_TMODES[1]:
                mem.rtone = TS790E_TONES[int(_p8)-1]
        if mem.number >= 50:
            mem.extd_number = TS790E_SPECIALS[mem.number - 50]
        return mem

    def _parse_split_spec(self, mem, spec):
        split_freq_bytes = spec[7:17]  # P4
        if len(split_freq_bytes) > 0:
            split_freq = int(split_freq_bytes)
            if mem.freq != 0:
                mem.duplex = "split"
                mem.offset = split_freq
        return mem

    def _make_base_spec(self, mem, freq):
        if mem.mode == "FM" \
                and mem.tmode == TS790E_TMODES[1]:
            tmode = "1"
            tone = "%02i" % (TS790E_TONES.index(mem.rtone)+1)
        else:
            tmode = "0"
            tone = "01"
        if not mem.offset or mem.offset not in TS790E_OFFSETS:
            offset = "0"
        else:
            offset = str(TS790E_OFFSETS.index(mem.offset))
        return "%011i%s%i%s%s%s" % (            # P# Format Function
                   freq,                        # P4 4      Frequency
                   TS790E_MODES[mem.mode],      # P5 2      Mode
                   mem.skip == TS790E_SKIP[1],  # P6 10     Memory Lockout
                   tmode,                       # P7 1      Tone On/Off
                   tone,                        # P8 14     Tone Frequency
                   offset)                      # P9 13     Offset



#!/usr/bin/env python3

import argparse
import contextlib
import logging
import random
import serial
import time
import sys

from chirp import bitwise
from chirp.drivers import icomciv
from chirp import memmap

LOG = logging.getLogger()

# Note these are only the frequencies from the chart that fit into the
# US band edges.
BANDS = {
    160: [i for i in range(1805, 2005, 20)],
    80: [i for i in range(3510, 4010, 10)],
    40: [i for i in range(7013, 7313, 25)],
    30: [10125],
    20: [i for i in range(14025, 14325, 50)],
    17: [18075, 18125, 18165],
    15: [i for i in range(21025, 21475, 50)],
    12: [24891, 24963],
    10: [i for i in range(28050, 29750, 100)],
    6: [i for i in range(50125, 54125, 250)],
}


class BitwiseFrame(icomciv.Frame):
    _fmt = ''
    _datalen = 0
    _querycmd: int | None = None

    def __init__(self):
        super().__init__()
        self._data = memmap.MemoryMapBytes(b'\x00' * self._datalen)
        self.parse()

    def parse(self):
        self._obj = bitwise.parse(self._fmt, self._data)

    @classmethod
    def get(cls, radio):
        f = icomciv.Frame()
        f.set_command(cls._querycmd or cls._cmd, cls._sub)
        radio._send_frame(f)
        r = radio._recv_frame(frame=cls())
        # If we used a different command to query, reset the received frame
        # to the one we'd use for the setting
        r._cmd = cls._cmd
        return r


class SetFreqFrame(BitwiseFrame):
    _datalen = 5
    _cmd = 0x05
    _querycmd = 0x03
    _sub = None
    _fmt = 'lbcd freq[5];'

    @property
    def freq(self):
        return self._obj.freq // 1000

    @freq.setter
    def freq(self, khz):
        self._obj.freq = khz * 1000


class SetModeFrame(BitwiseFrame):
    _datalen = 2
    _cmd = 0x06
    _querycmd = 0x04
    _sub = None
    _fmt = 'u8 mode; u8 filter;'
    _modes = [
        ('LSB', 0),
        ('USB', 1),
        ('AM', 2),
        ('CW', 3),
        ('RTTY', 4),
        ('FM', 5),
        ('CW-R', 7),
        ('RTTY-R', 8),
        ('PSK', 12),
        ('PSK-R', 17),
    ]

    @property
    def mode(self):
        for mode, index in self._modes:
            if index == self._obj.mode:
                return mode
        raise RuntimeError('Unsupported mode %i' % self._obj.mode)

    @mode.setter
    def mode(self, mode):
        self._obj.filter = 1
        for m, index in self._modes:
            if mode == m:
                self._obj.mode = index
                return
        raise RuntimeError('Unsupported mode %s' % mode)

    @property
    def filter(self):
        return int(self._obj.filter)


class BKINFrame(BitwiseFrame):
    _datalen = 1
    _cmd = 0x16
    _sub = 0x47
    _fmt = 'u8 mode;'
    _modes = ('off', 'semi', 'full')

    @property
    def mode(self):
        return self._modes[self._obj.mode]

    @mode.setter
    def mode(self, mode):
        self._obj.mode = self._modes.index(mode)


class CWSpeedFrame(BitwiseFrame):
    _datalen = 2
    _cmd = 0x14
    _sub = 0x0C
    _fmt = 'bbcd speed[2];'

    @property
    def speed(self):
        return 48 * self._obj.speed // 255

    @speed.setter
    def speed(self, wpm):
        self._obj.speed = int((wpm / 48) * 255)


class PowerFrame(BitwiseFrame):
    _datalen = 2
    _cmd = 0x15
    _sub = 0x11
    _fmt = 'bbcd power[2];'

    @classmethod
    def get(cls, radio):
        f = icomciv.Frame()
        f.set_command(cls._cmd, cls._sub)
        radio._send_frame(f)
        f = radio._recv_frame(frame=cls())
        f.parse()
        return f

    @property
    def power(self):
        return int(100 * int(self._obj.power) / 212)

    @power.setter
    def power(self, power):
        self._obj.power = 212 * (power / 100)


class Demo:
    def __init__(self, s=None):
        self._last_command = 0

    def _send_frame(self, f):
        self._last_command = f._cmd

    def _recv_frame(self, frame=None):
        if frame is None:
            frame = PowerFrame()
        if self._last_command == 0x15:
            frame.power = random.randint(0, 10) * 10
        else:
            frame.set_command(0xFB, None)
        return frame


def setup(port, rcls):
    LOG.debug('Connecting to %s' % port)
    s = serial.Serial(port, timeout=1)
    LOG.debug('Starting radio')
    radio = rcls(s)
    LOG.debug('Detecting baudrate')
    radio._detect_baudrate()
    LOG.debug('Connected')
    return radio


def set_freq(radio, khz, mode):
    f = SetFreqFrame()
    f.freq = khz
    radio._send_frame(f)
    r = radio._recv_frame()
    assert r._cmd == 0xFB
    LOG.debug('Set frequency to %i kHz', khz)

    f = SetModeFrame()
    f.mode = mode
    radio._send_frame(f)
    r = radio._recv_frame()
    assert r._cmd == 0xFB
    LOG.debug('Set mode to %s', mode)


def interactive_next(current):
    print()
    print('Options:')
    print('r: Retry this frequency')
    print('x: Exit immediately')
    print('[enter]: Continue to the next frequency')
    i = input('> ').lower()
    if i == '':
        return True
    elif i == 'r':
        return False
    elif i == 'x':
        raise RuntimeError('Abort')


def auto_next(current):
    print()
    return True


def send_id(radio, call):
    f = SetModeFrame()
    f.mode = 'CW'
    radio._send_frame(f)
    r = radio._recv_frame()
    assert r._cmd == 0xFB
    LOG.debug('Set mode to CW')

    f = BKINFrame()
    f.mode = 'full'
    radio._send_frame(f)
    r = radio._recv_frame()
    assert r._cmd == 0xFB
    LOG.debug('Set BK-IN to full')

    f = CWSpeedFrame()
    f.speed = 22
    radio._send_frame(f)
    r = radio._recv_frame()
    assert r._cmd == 0xFB
    LOG.debug('Set key speed to 22WPM')

    f = icomciv.Frame()
    f.set_command(0x17, 0x40)
    f.set_data(b'DE %s' % call.upper().encode())
    radio._send_frame(f)
    r = radio._recv_frame()
    if r._cmd != 0xFB:
        print('Radio does not support CWID - ID manually now')
    else:
        print('Sending id %s' % call)
    time.sleep(len(call) + 2)
    LOG.debug('Sent ID')


def wait_for_power(radio, fn):
    po = None
    maxl = 0
    while po is None or not fn(po):
        po = PowerFrame.get(radio).power
        line = sys.stdout.write('Po: |%-10s| %3iW\r' % ('=' * (po // 10), po))
        maxl = max(maxl, line)
        time.sleep(0.5)
    sys.stdout.write('%s\r' % (' ' * maxl))


def wait_for_tune(radio):
    print('Transmit to tune now (waiting for key)')
    wait_for_power(radio, lambda po: po > 0)
    print('Keyed, waiting for completion')
    wait_for_power(radio, lambda po: po == 0)
    print('Done')


@contextlib.contextmanager
def restore_settings(radio):
    if not isinstance(radio, Demo):
        settings = [
            SetModeFrame.get(radio),
            SetFreqFrame.get(radio),
            BKINFrame.get(radio),
            CWSpeedFrame.get(radio),
        ]
    else:
        settings = []
    try:
        yield
    finally:
        LOG.debug('Restoring previous settings')
        for f in settings:
            radio._send_frame(f)
            radio._recv_frame()


def tune_loop(radio, bands, call, next_fn):
    for band in bands:
        for freq in BANDS[band]:
            done = False
            while not done:
                print('Changing to %i kHz' % freq)
                set_freq(radio, freq, 'RTTY')
                time.sleep(1)

                wait_for_tune(radio)
                time.sleep(2)
                if call:
                    send_id(radio, call)

                try:
                    done = next_fn(freq)
                except RuntimeError:
                    return 0


def main():
    # More of these may work, but these are known to work
    radios = {
        '7200': icomciv.Icom7200Radio,
        '7300': icomciv.Icom7300Radio,
        '7610': icomciv.Icom7610Radio,
        'Demo': Demo,
    }

    p = argparse.ArgumentParser(description=(
        'A simple tool to automate running through the required frequencies '
        'to tune an SPE Expert linear.'))
    p.add_argument('radio', choices=radios,
                   help='Radio model')
    p.add_argument('port', help='Serial port for CAT control')
    p.add_argument('--bands', help=('Comma-separated list of bands '
                                    'to tune (160, 80, etc)'))
    p.add_argument('--call', help='Callsign for CW ID after each step')
    p.add_argument('--next',
                   choices=['interactive', 'auto'],
                   default='interactive',
                   help='Next step strategy')
    p.add_argument('--debug', action='store_true',
                   help='Enable verbose debugging')
    args = p.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    next_fn = globals()['%s_next' % args.next]

    try:
        bands = [int(i) for i in args.bands.split(',')]
        for band in bands:
            assert band in BANDS
    except (ValueError, AssertionError):
        print('Invalid band input. Must be one or more of %s' % (
            ','.join(str(i) for i in BANDS.keys())))
        return 1

    if args.radio == 'Demo':
        radio = Demo()
    else:
        radio = setup(args.port, radios[args.radio])

    with restore_settings(radio):
        tune_loop(radio, bands, args.call, next_fn)


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



from chirp import util
from tests.unit import base


class TestUtils(base.BaseTest):
    def test_hexprint_with_string(self):
        util.hexprint('00000000000000')

    def test_hexprint_with_bytes(self):
        util.hexprint(b'00000000000000')

    def test_hexprint_short(self):
        expected = ('000: 00 00 00 00 00 00 00 00   ........\n'
                    '008: 00                        ........\n')
        self.assertEqual(expected, util.hexprint(b'\x00' * 9))

    def test_hexprint_even(self):
        expected = '000: 00 00 00 00 00 00 00 00   ........\n'
        self.assertEqual(expected, util.hexprint(b'\x00' * 8))



import os
import unittest

from chirp.drivers import th9800


class Fake9800:
    ident = (b'\x54\x48\x39\x38\x30\x30\xff\xff'
             b'\x56\x31\x34\x78\xff\xff\xff\xff')

    def __init__(self):
        self.writebuf = b''
        # Pre-load the identify dance
        self.readbuf = b'A' + self.ident

    def write(self, data):
        assert isinstance(data, bytes)
        self.writebuf += data
        if data.startswith(b'R'):
            self.readbuf += b'W\x00\x00\x00' + (b'\x00' * 0x80)
        elif data.startswith(b'W'):
            self.writebuf += data
            self.readbuf += b'A'
        elif data == b'A':
            assert len(self.readbuf) == 0
            self.readbuf += b'A'
        elif data == b'ENDW':
            self.readbuf += b'123'

    def read(self, n):
        buf = self.readbuf[:n]
        self.readbuf = self.readbuf[n:]
        return buf


class TestTH9800(unittest.TestCase):
    def test_identify(self):
        f = Fake9800()
        r = th9800.TYTTH9800Radio(f)
        ident = th9800._identify(r)
        self.assertEqual(f.ident, ident)
        self.assertEqual(b'\x02PROGRAM\x02A', f.writebuf)

    def test_download(self):
        f = Fake9800()
        r = th9800.TYTTH9800Radio(f)
        r.sync_in()

    def test_upload(self):
        f = Fake9800()
        img = os.path.join(os.path.dirname(__file__),
                           '..', 'images', 'TYT_TH-9800.img')
        r = th9800.TYTTH9800Radio(img)
        r.set_pipe(f)
        r.sync_out()



# Copyright 2025 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct memory {
    u32 rxfreq;      // 0-3
    ul16 rxtone;     // 4-5
    u32 txfreq;      // 6-9
    ul16 txtone;     // A-B
    u8 unknown0:3,   // C
       scan:1,       //
       is_wide:1,    //
       unknown1:1,   //
       highpower:1,  //
       unknown2:1;   //
    u8 unknown3[3];  // D-F
} ;

struct memory lochannels[14];
struct memory chnl15;
struct memory rptr15;
struct memory chnl16;
struct memory rptr16;
struct memory chnl17;
struct memory rptr17;
struct memory chnl18;
struct memory rptr18;
struct memory chnl19;
struct memory rptr19;
struct memory chnl20;
struct memory rptr20;
struct memory chnl21;
struct memory rptr21;
struct memory chnl22;
struct memory rptr22;
struct memory hichannels[77];

#seekto 0x0E00;
struct {
    u8 unknown00:4,   vox:4;         // 0E00 VOX
    u8 unknown01:7,   keytone:1;     // 0E01 Key Tone
    u8 unknown02:7,   roger:1;       // 0E02 Roger Beep
    u8 unknown03:5,   calltone:3;    // 0E03 Call Tone
    u8 unknown04:6,   brightness:2;  // 0E04 Screen Brightness
    u8 unknown05:5,   backlight:3;   // 0E05 Backlight Duration
    u8 unknown06:4,   squelch:4;     // 0E06 Squelch Level
    u8 unknown07;                    // 0E07 (Reserved/Unknown)
    u8 unknown08:7,   repeat:1;      // 0E08 Repeater
    u8 unknown09:5,   sleeptime:3;   // 0E09 Sleep Timer
    u8 unknown10:6,   speakersw:2;   // 0E0A Speaker Switch
    u8 unknown11:6,   micgain:2;     // 0E0B Mic Gain
    u8 unknown12:4,   tot:4;         // 0E0C Timeout Timer (TOT)
    u8 unknown13;                    // 0E0D (Reserved)
    u8 unknown14;                    // 0E0E (Reserved)
    u8 unknown15;                    // 0E0F (Reserved)
} settings;
"""

CMD_ACK = b"\x06"

LIST_OFF1TO9 = ["Off"] + ["%s" % x for x in range(1, 10)]
LIST_SLEEPTIME = ["OFF", "5s", "10s", "20s", "30s", "60s"]
LIST_SPEAKERSW = ["Base", "Mic", "Dual"]
LIST_TOT = ["Off"] + ["%s" % x for x in range(15, 195, 15)]

SPECIALS = {
        "RPTR15": -8,
        "RPTR16": -7,
        "RPTR17": -6,
        "RPTR18": -5,
        "RPTR19": -4,
        "RPTR20": -3,
        "RPTR21": -2,
        "RPTR22": -1
        }


def _enter_programming_mode(radio):
    serial = radio.pipe

    _magic = b"\x02" + radio._magic

    try:
        serial.write(_magic)
        serial.read(len(_magic))  # Chew the echo
        ack = serial.read(1)
    except Exception:
        raise errors.RadioError("Error communicating with radio")

    if not ack:
        raise errors.RadioError("No response from radio")
    elif ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")

    try:
        serial.write(b"\x02")
        serial.read(1)  # Chew the echo
        ident = serial.read(8)
    except Exception:
        raise errors.RadioError("Error communicating with radio")

    # check if ident is OK
    for fp in radio._fingerprint:
        if ident.startswith(fp):
            break
    else:
        LOG.debug("Incorrect model ID, got this:\n\n" + util.hexprint(ident))
        raise errors.RadioError("Radio identification failed.")

    try:
        serial.write(CMD_ACK)
        serial.read(1)  # Chew the echo
        ack = serial.read(1)
    except Exception:
        raise errors.RadioError("Error communicating with radio")

    if ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")


def _exit_programming_mode(radio):
    serial = radio.pipe
    try:
        serial.write(radio.CMD_EXIT)
        serial.read(1)  # Chew the echo
    except Exception:
        raise errors.RadioError("Radio refused to exit programming mode")


def _read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', block_addr, block_size)
    expectedresponse = b"R" + cmd[1:]
    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        serial.read(4)  # Chew the echo
        response = serial.read(4 + block_size)
        if response[:4] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        block_data = response[4:]

    except Exception:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    return block_data


def _write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'W', block_addr, block_size)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        serial.read(4 + len(data))  # Chew the echo
        serial.read(1)  # Chew the echo
        ack = serial.read(1)
        if ack != CMD_ACK:
            raise Exception("No ACK")
    except Exception:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")
    _enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    _exit_programming_mode(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE):
            status.cur = addr + radio.BLOCK_SIZE
            radio.status_fn(status)
            _write_block(radio, addr, radio.BLOCK_SIZE)

    _exit_programming_mode(radio)


@directory.register
class RA86Radio(chirp_common.CloneModeRadio):
    """RETEVIS RA86"""
    VENDOR = "Retevis"
    MODEL = "RA86"
    BAUD_RATE = 115200
    BLOCK_SIZE = 0x10
    CMD_EXIT = b"b"

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=20.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    VALID_BANDS = [(400000000, 470000000)]

    DTCS_CODES = tuple(sorted(chirp_common.DTCS_CODES + (645,)))

    _magic = b"PROGRAM"
    _fingerprint = [b"SMP558"]
    _upper = 99

    _gmrs = False

    _ranges = [
        (0x0000, 0x06B0),
        (0x0E00, 0x0E10),
    ]
    _memsize = 0x0E10

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.valid_special_chans = list(SPECIALS.keys())
        rf.has_settings = True
        rf.valid_modes = ["FM", "NFM"]  # 25 kHz, 12.5 kHz.
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.can_odd_split = True
        rf.has_rx_dtcs = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_dtcs_codes = self.DTCS_CODES
        rf.has_tuning_step = False
        rf.has_bank = False
        rf.has_name = False
        rf.memory_bounds = (1, self._upper)
        rf.valid_bands = self.VALID_BANDS
        rf.valid_tuning_steps = chirp_common.TUNING_STEPS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    def sync_out(self):
        do_upload(self)

    def decode_tone(self, val):
        mode = ""
        pol = "N"
        if val in [0, 0xFFFF]:
            xval = 0
        elif val <= 0x32:
            mode = "Tone"
            index = val - 1
            xval = chirp_common.TONES[index]
        elif val >= 0x0033:
            mode = "DTCS"
            if val > 0x69 + 0x32:
                index = val - 0x69 - 0x32 - 1
                pol = "R"
            else:
                index = val - 0x32 - 1
            xval = self.DTCS_CODES[index]
        else:
            LOG.warn("Bug: tone is %04x" % val)
        return mode, xval, pol

    def encode_tone(self, memtone, mode, tone, pol):
        if mode == "Tone":
            memtone.set_value(chirp_common.TONES.index(tone) + 1)
        elif mode == "TSQL":
            memtone.set_value(chirp_common.TONES.index(tone) + 1)
        elif mode == "DTCS":
            if pol == 'R':
                memtone.set_value(self.DTCS_CODES.index(tone) + 0x33 + 0x69)
            else:
                memtone.set_value(self.DTCS_CODES.index(tone) + 0x33)
        else:
            memtone.set_value(0)

    def _get_memobjs(self, number):
        if isinstance(number, str):
            return (getattr(self._memobj, number.lower()))
        elif number < 0:
            for k, v in SPECIALS.items():
                if number == v:
                    return (getattr(self._memobj, k.lower()))
        else:
            if number < 15:
                return self._memobj.lochannels[number - 1]
            elif 15 <= number <= 22:
                return getattr(self._memobj, f"chnl{number}")
            else:  # number > 22
                return self._memobj.hichannels[number - 23]

    def get_memory(self, number):
        _mem = self._get_memobjs(number)

        mem = chirp_common.Memory()

        if isinstance(number, str):
            mem.number = SPECIALS[number]
            mem.extd_number = number
        else:
            mem.number = number

        mem.freq = int(_mem.rxfreq)

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True
            return mem

        if _mem.rxfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.freq = 0
            mem.empty = True
            return mem

        if _mem.get_raw() == (b"\xFF" * 16):
            LOG.debug("Initializing empty memory")
            _mem.set_raw(b"\x00" * 16)

        # Freq and offset
        mem.freq = int(_mem.rxfreq)
        # tx freq can be blank
        if _mem.txfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            # TX freq not set
            mem.offset = 0
            mem.duplex = "off"
        else:
            # TX freq set
            offset = (int(_mem.txfreq)) - mem.freq
            if offset != 0:
                if chirp_common.is_split(self.get_features().valid_bands,
                                         mem.freq, int(_mem.txfreq)):
                    mem.duplex = "split"
                    mem.offset = int(_mem.txfreq)
                elif offset < 0:
                    mem.offset = abs(offset)
                    mem.duplex = "-"
                elif offset > 0:
                    mem.offset = offset
                    mem.duplex = "+"
            else:
                mem.offset = 0

        mem.mode = _mem.is_wide and "FM" or "NFM"

        if not _mem.scan:
            mem.skip = "S"

        txtone = self.decode_tone(_mem.txtone)
        rxtone = self.decode_tone(_mem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        mem.power = self.POWER_LEVELS[1 - _mem.highpower]

        return mem

    def set_memory(self, mem):
        # Get a low-level memory object mapped to the image
        _mem = self._get_memobjs(mem.number)

        if mem.empty:
            _mem.fill_raw(b"\xFF")

            return

        _mem.fill_raw(b"\x00")

        _mem.rxfreq = mem.freq

        if mem.duplex == "off":
            _mem.txfreq.fill_raw(b"\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset)
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset)
        else:
            _mem.txfreq = mem.freq

        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self.encode_tone(_mem.txtone, txmode, txtone, txpol)
        self.encode_tone(_mem.rxtone, rxmode, rxtone, rxpol)

        _mem.highpower = mem.power == self.POWER_LEVELS[0]

        _mem.is_wide = mem.mode == "FM"
        _mem.scan = mem.skip != "S"

        _mem.unknown0 = 0
        _mem.unknown1 = 0
        _mem.unknown2 = 0
        _mem.unknown3 = b"\x92\x1B\xC4"

    def get_settings(self):
        _mem = self._memobj
        basic = RadioSettingGroup("basic", "Basic Settings")
        top = RadioSettings(basic)

        if _mem.settings.vox >= len(LIST_OFF1TO9):
            val = 0x00
        else:
            val = _mem.settings.vox
        rs = RadioSetting("settings.vox", "VOX",
                          RadioSettingValueList(
                              LIST_OFF1TO9,
                              current_index=val))
        basic.append(rs)

        rs = RadioSetting("settings.keytone", "Key Tone",
                          RadioSettingValueBoolean(_mem.settings.keytone))
        basic.append(rs)

        rs = RadioSetting("settings.roger", "Roger Beep",
                          RadioSettingValueBoolean(_mem.settings.roger))
        basic.append(rs)

        if _mem.settings.calltone > 4:
            val = 1
        else:
            val = _mem.settings.calltone + 1
        rs = RadioSetting("settings.calltone", "Call Tone",
                          RadioSettingValueInteger(1, 5, val))
        basic.append(rs)

        if _mem.settings.brightness > 2:
            val = 1
        else:
            val = _mem.settings.brightness + 1
        rs = RadioSetting("settings.brightness", "Brightness",
                          RadioSettingValueInteger(1, 3, val))
        basic.append(rs)

        if _mem.settings.backlight > 4:
            val = 2
        else:
            val = _mem.settings.backlight
        rs = RadioSetting("settings.backlight", "Back Light",
                          RadioSettingValueInteger(0, 4, val))
        basic.append(rs)

        if _mem.settings.squelch >= len(LIST_OFF1TO9):
            val = 0x05
        else:
            val = _mem.settings.squelch
        rs = RadioSetting("settings.squelch", "Squelch",
                          RadioSettingValueList(
                              LIST_OFF1TO9,
                              current_index=val))
        basic.append(rs)

        if _mem.settings.tot >= len(LIST_TOT):
            val = 0x0C
        else:
            val = _mem.settings.tot
        rs = RadioSetting("settings.tot", "TOT",
                          RadioSettingValueList(
                              LIST_TOT,
                              current_index=val))
        basic.append(rs)

        rs = RadioSetting("settings.repeat", "Repeater",
                          RadioSettingValueBoolean(_mem.settings.repeat))
        basic.append(rs)

        if _mem.settings.sleeptime >= len(LIST_SLEEPTIME):
            val = 0x02
        else:
            val = _mem.settings.sleeptime
        rs = RadioSetting("settings.sleeptime", "Sleep Time",
                          RadioSettingValueList(
                              LIST_SLEEPTIME,
                              current_index=val))
        basic.append(rs)

        if _mem.settings.speakersw >= len(LIST_SPEAKERSW):
            val = 0x02
        else:
            val = _mem.settings.speakersw
        rs = RadioSetting("settings.speakersw", "Speaker Select",
                          RadioSettingValueList(
                              LIST_SPEAKERSW,
                              current_index=val))
        basic.append(rs)

        if _mem.settings.micgain > 2:
            val = 1
        else:
            val = _mem.settings.micgain + 1
        rs = RadioSetting("settings.micgain", "Mic Gain",
                          RadioSettingValueInteger(1, 5, val))
        basic.append(rs)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "brightness":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "calltone":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "micgain":
                        setattr(obj, setting, int(element.value) - 1)
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise



from builtins import bytes
import os
import tempfile
import unittest

from chirp import directory
from chirp.drivers import ft60
from chirp.drivers import ft4
from chirp.drivers import yaesu_clone
from chirp import memmap


class TestYaesuChecksum(unittest.TestCase):
    def _test_checksum(self, mmap):
        cs = yaesu_clone.YaesuChecksum(0, 2, 3)

        self.assertEqual(42, cs.get_existing(mmap))
        self.assertEqual(0x8A, cs.get_calculated(mmap))
        try:
            mmap = mmap.get_byte_compatible()
            mmap[0] = 3
        except AttributeError:
            # str or bytes
            try:
                # str
                mmap = memmap.MemoryMap('\x03' + mmap[1:])
            except TypeError:
                # bytes
                mmap = memmap.MemoryMapBytes(b'\x03' + mmap[1:])

        cs.update(mmap)
        self.assertEqual(95, cs.get_calculated(mmap))

    def test_with_MemoryMap(self):
        mmap = memmap.MemoryMap('...\x2A')
        self._test_checksum(mmap)

    def test_with_MemoryMapBytes(self):
        mmap = memmap.MemoryMapBytes(bytes(b'...\x2A'))
        self._test_checksum(mmap)

    def test_with_bytes(self):
        self._test_checksum(b'...\x2A')

    def test_with_str(self):
        self._test_checksum('...\x2A')


class FakeFT60:
    def __init__(self):
        self.readbuf = b''
        self.writebuf = b''

    def start_download(self):
        self.readbuf += b'\x00' * 8
        for i in range(448):
            self.readbuf += b'\x00' * 64

    def write(self, data):
        assert isinstance(data, bytes)
        # Prepend echo so it is read first next
        self.readbuf = data + self.readbuf

        self.writebuf += data
        if len(data) == 1:
            # If we're doing an upload, we need to ack this
            # last short block
            if self.writebuf.startswith(b'AH017'):
                self.readbuf += b'\x06'
        elif len(data) in (8, 64):
            self.readbuf += b'\x06'
        else:
            raise Exception('Unhandled')

    def read(self, n):
        buf = self.readbuf[:n]
        self.readbuf = self.readbuf[n:]
        return buf


class TestFT60(unittest.TestCase):
    def test_download(self):
        f = FakeFT60()
        f.start_download()
        r = ft60.FT60Radio(f)
        r.sync_in()
        # Make sure the ACKs are there
        self.assertEqual(449, len(f.writebuf))
        self.assertEqual(f.writebuf, b'\x06' * 449)

    def test_upload(self):
        f = FakeFT60()
        img = os.path.join(os.path.dirname(__file__),
                           '..', 'images', 'Yaesu_FT-60.img')
        r = ft60.FT60Radio(img)
        r.set_pipe(f)
        r.sync_out()


class FakeFTX4:
    ID = ft4.YaesuFT4XERadio.id_str

    def __init__(self):
        self.readbuf = b''
        self.writebuf = b''

    def write(self, buf):
        assert isinstance(buf, bytes)
        # Echo
        self.readbuf = buf + self.readbuf
        if buf.startswith(b'PROGRAM'):
            self.readbuf += b'QX'
        elif buf == b'\x02':
            # Ident
            self.readbuf += self.ID
        elif buf.startswith(b'R'):
            resp = b'W' + buf[1:] + b'\x00' * 16
            self.readbuf += resp + bytes([ft4.checkSum8(resp[1:])])
        elif buf.startswith(b'W'):
            pass
        elif buf == b'END':
            pass
        else:
            raise Exception('Unhandled %r' % buf)
        self.readbuf += b'\x06'

    def read(self, n):
        buf = self.readbuf[:n]
        self.readbuf = self.readbuf[n:]
        return buf


class FakeFT25R(FakeFTX4):
    ID = ft4.YaesuFT25RRadio.id_str


class FakeFT25R_Asian(FakeFTX4):
    ID = ft4.YaesuFT25RRadio.id_str[:-1] + b'\x03'


class TestFTX4(unittest.TestCase):
    RCLASS = ft4.YaesuFT4XERadio
    FAKE = FakeFTX4

    def test_download(self):
        f = self.FAKE()
        r = self.RCLASS(f)
        r.sync_in()
        self.assertEqual(f.ID[-1], r.subtype)

    def test_upload(self):
        f = self.FAKE()
        fn = directory.radio_class_id(self.RCLASS)
        img = os.path.join(os.path.dirname(__file__),
                           '..', 'images', '%s.img' % fn)
        r = self.RCLASS(img)
        r.set_pipe(f)
        r.sync_out()

    def test_download_open(self):
        f = self.FAKE()
        r = self.RCLASS(f)
        r.sync_in()
        # Make sure we got subtype set as expected after download
        self.assertEqual(f.ID[-1], r.subtype)

        # Save it out to a file
        fn = tempfile.mktemp('.img', 'ft4')
        r.save(fn)

        # Make sure if we re-load our image, we keep the same subtype
        r = self.RCLASS(fn)
        self.assertEqual(f.ID[-1], r.subtype)


class TestFT25(TestFTX4):
    RCLASS = ft4.YaesuFT25RRadio
    FAKE = FakeFT25R


class TestFT25_Asian(TestFT25):
    FAKE = FakeFT25R_Asian



import pytest
from unittest import mock

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp import errors
from tests import base


class TestCaseEdges(base.DriverTest):
    def test_inherits_from_registered(self):
        if self.rf.has_sub_devices:
            devs = self.radio.get_sub_devices()
        else:
            devs = [self.radio]
        for dev in devs:
            directory.registered_class(dev.__class__)

    def test_longname(self):
        m = self.get_mem()
        m.name = ("X" * 256)  # Should be longer than any radio can handle
        m.name = self.radio.filter_name(m.name)

        self.radio.set_memory(m)
        n = self.radio.get_memory(m.number)

        self.assertEqualMem(m, n)

    def test_badname(self):
        m = self.get_mem()

        ascii = "".join([chr(x) for x in range(ord(" "), ord("~")+1)])
        for i in range(0, len(ascii), 4):
            m.name = self.radio.filter_name(ascii[i:i+4])
            self.radio.set_memory(m)
            n = self.radio.get_memory(m.number)
            self.assertEqualMem(m, n)

    def test_bandedges(self):
        m = self.get_mem()
        min_step = min(self.rf.has_tuning_step and
                       self.rf.valid_tuning_steps or [10])

        for low, high in self.rf.valid_bands:
            for freq in (low, high - int(min_step * 1000)):
                try:
                    m.freq = freq
                except chirp_common.ImmutableValueError:
                    self.skipTest('Test memory has immutable freq')
                if self.radio.validate_memory(m):
                    # Radio doesn't like it, so skip
                    continue

                self.radio.set_memory(m)
                n = self.radio.get_memory(m.number)
                self.assertEqualMem(m, n)

    def test_oddsteps(self):
        odd_steps = {
            145000000: [145856250, 145862500],
            445000000: [445856250, 445862500],
            862000000: [862731250, 862737500],
        }

        m = self.get_mem()

        for low, high in self.rf.valid_bands:
            for band, totest in list(odd_steps.items()):
                if band < low or band > high:
                    continue
                for testfreq in totest:
                    step = chirp_common.required_step(testfreq)
                    if step not in self.rf.valid_tuning_steps:
                        continue

                    try:
                        m.freq = testfreq
                    except chirp_common.ImmutableValueError:
                        self.skipTest('Test channel has immutable freq')
                    m.tuning_step = step
                    self.radio.set_memory(m)
                    n = self.radio.get_memory(m.number)
                    # Some radios have per-band required modes, which we
                    # don't care about testing here
                    self.assertEqualMem(m, n, ignore=['tuning_step',
                                                      'mode'])

    def test_empty_to_not(self):
        firstband = self.rf.valid_bands[0]
        testfreq = firstband[0]
        for loc in range(*self.rf.memory_bounds):
            m = self.radio.get_memory(loc)
            if m.empty:
                m.empty = False
                m.freq = testfreq
                self.radio.set_memory(m)
                m = self.radio.get_memory(loc)
                self.assertEqual(testfreq, m.freq,
                                 'Radio returned an unexpected frequency when'
                                 'setting previously-empty location %i' % loc)
                break

    def test_delete_memory(self):
        for loc in range(*self.rf.memory_bounds):
            m = self.radio.get_memory(loc)
            if 'empty' in m.immutable:
                # This memory is not deletable
                continue
            if not m.empty:
                m.empty = True
                self.radio.set_memory(m)
                m = self.radio.get_memory(loc)
                self.assertTrue(m.empty,
                                'Radio returned non-empty memory when asked '
                                'to delete location %i' % loc)
                break

    def test_redelete_memory(self):
        m = self.get_mem()
        if 'empty' in m.immutable:
            self.skipTest('Test memory is not deletable')

        # Delete this memory
        m.empty = True
        self.radio.set_memory(m)

        # Try to delete it again
        self.radio.set_memory(m)

        # Make sure it appears deleted
        m2 = self.radio.get_memory(m.number)
        self.assertTrue(m2.empty)

    def test_get_set_specials(self):
        if not self.rf.valid_special_chans:
            self.skipTest('Radio has no specials')
        lo, hi = self.rf.memory_bounds
        for name in self.rf.valid_special_chans:
            m1 = self.radio.get_memory(name)
            # Flip to non-empty, but only touch it if it's false because some
            # radios have empty in the immutable set.
            if m1.empty:
                m1.empty = False

            self.assertIsInstance(m1.number, int,
                                  ('Special memory %s number %r is not an '
                                   'integer') % (name, m1.number))

            try:
                del m1.extra
            except AttributeError:
                pass
            if m1.freq > 130000000 and m1.freq < 500000000:
                m1.freq += 5000
            elif m1.freq == 0:
                # If the memory was empty before, we likely need to pick
                # a valid frequency. Use the bottom of the first supported
                # band.
                m1.freq = self.rf.valid_bands[0][0]
            try:
                self.radio.set_memory(m1)
            except errors.RadioError:
                # If the radio raises RadioError, assume we're editing a
                # special channel that is not editable
                continue
            m2 = self.radio.get_memory(name)
            self.assertEqualMem(m1, m2, ignore=['name'])

            self.assertFalse(
                lo < m1.number < hi,
                'Special memory %s maps into memory bounds at %i' % (
                    name, m1.number))

    def test_check_regular_not_special(self):
        lo, hi = self.rf.memory_bounds
        for i in range(lo, hi + 1):
            m = self.radio.get_memory(i)
            self.assertEqual('', m.extd_number,
                             'Non-special memory %i should not have '
                             'extd_number set to %r' % (i, m.extd_number))
            self.assertIsInstance(m.number, int,
                                  'Memory number %r is not an integer' %
                                  m.number)

    def test_get_memory_name_trailing_whitespace(self):
        if self.radio.MODEL == 'KG-UV8E':
            self.skipTest('The UV8E driver is so broken it does not even '
                          'parse its own test image cleanly')
        for i in range(*self.rf.memory_bounds):
            m = self.radio.get_memory(i)
            self.assertEqual(m.name.rstrip(), m.name,
                             'Radio returned a memory with trailing '
                             'whitespace in the name')

    def test_memory_name_stripped(self):
        if ' ' not in self.rf.valid_characters:
            self.skipTest('Radio does not support space characters')
        m1 = self.get_mem()
        m1.name = 'FOO '
        self.radio.set_memory(m1)
        m2 = self.radio.get_memory(m1.number)
        self.assertEqual(m2.name.rstrip(), m2.name,
                         'Radio set and returned a memory with trailing '
                         'whitespace in the name')


class TestBitwiseStrict(base.DriverTest):
    def setUp(self):
        pass

    def _raise(self, message):
        raise SyntaxError(message)

    @pytest.mark.xfail(strict=False)
    @mock.patch.object(bitwise.Processor, 'assert_negative_seek',
                       side_effect=_raise)
    def test_bitwise_negative_seek(self, mock_assert):
        super().setUp()

    @pytest.mark.xfail(strict=False)
    @mock.patch.object(bitwise.Processor, 'assert_unnecessary_seek',
                       side_effect=_raise)
    def test_bitwise_unnecessary_seek(self, mock_assert):
        super().setUp()

    @mock.patch.object(bitwise.LOG, 'error')
    def test_bitwise_errors(self, mock_log):
        super().setUp()
        self.assertFalse(mock_log.called,
                         [x[0][0] for x in mock_log.call_args_list])



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum


class Reasons(enum.Enum):
    NO_CONNECTION_K1 = "No response from radio. Check connector and cabling!"
    FIXED_BANKS = "This radio has fixed banks and does not allow reassignment"


class InvalidDataError(Exception):
    """The radio driver encountered some invalid data"""
    pass


class InvalidValueError(Exception):
    """An invalid value for a given parameter was used"""
    pass


class InvalidMemoryLocation(Exception):
    """The requested memory location does not exist"""
    pass


class RadioError(Exception):
    """An error occurred while talking to the radio"""
    pass


class UnsupportedToneError(Exception):
    """The radio does not support the specified tone value"""
    pass


class ImageDetectFailed(Exception):
    """The driver for the supplied image could not be determined"""
    pass


class ImageMetadataInvalidModel(Exception):
    """The image contains metadata but no suitable driver is found"""
    pass


class SpecificRadioError(RadioError):
    """An error with a specific reason and troubleshooting reference."""
    CODE: None | Reasons = None

    def __init__(self, msg=None):
        if self.CODE not in Reasons:
            raise RuntimeError('Invalid reason; '
                               'must be one of chirp.errors.Reasons')
        super().__init__(msg or self.CODE.value)

    def get_link(self):
        return ('https://chirpmyradio.com/projects/chirp/wiki/'
                'Error-%s' % self.CODE.name)


class RadioNoContactLikelyK1(SpecificRadioError):
    """A radio that uses a K1 connector likely to have fitment issues."""
    CODE = Reasons.NO_CONNECTION_K1


class RadioFixedBanks(SpecificRadioError):
    """A radio that has fixed banks and cannot be changed."""
    CODE = Reasons.FIXED_BANKS



from unittest import mock

from tests.unit import base
from chirp import import_logic
from chirp import chirp_common
from chirp import errors


class FakeRadio(chirp_common.Radio):
    def __init__(self, arg):
        self.POWER_LEVELS = list([chirp_common.PowerLevel('lo', watts=5),
                                  chirp_common.PowerLevel('hi', watts=50)])
        self.TMODES = list(['', 'Tone', 'TSQL'])
        self.HAS_CTONE = True
        self.HAS_RX_DTCS = False
        self.MODES = list(['FM', 'AM', 'DV'])
        self.DUPLEXES = list(['', '-', '+'])

    def filter_name(self, name):
        return 'filtered-name'

    def get_memory(self, number):
        return chirp_common.Memory(number, empty=True)

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.valid_bands = [(100000000, 500000000)]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_tmodes = self.TMODES
        rf.valid_modes = self.MODES
        rf.valid_duplexes = self.DUPLEXES
        rf.has_ctone = self.HAS_CTONE
        rf.has_rx_dtcs = self.HAS_RX_DTCS
        return rf


class FakeDstarRadio(FakeRadio, chirp_common.IcomDstarSupport):
    REQUIRES_CALL_LISTS = False

    def get_features(self):
        rf = super().get_features()
        rf.requires_call_lists = self.REQUIRES_CALL_LISTS
        return rf


class DstarTests(base.BaseTest):
    def _test_ensure_has_calls(self, mem,
                               ini_urcalls, ini_rptcalls,
                               exp_urcalls, exp_rptcalls):
        radio = FakeDstarRadio(None)
        with mock.patch.multiple(
                radio,
                get_urcall_list=mock.DEFAULT,
                get_repeater_call_list=mock.DEFAULT) as mocks:
            mocks['get_urcall_list'].return_value = ini_urcalls
            mocks['get_repeater_call_list'].return_value = ini_rptcalls
            import_logic.ensure_has_calls(radio, mem)
            self.assertEqual(sorted(ini_urcalls), sorted(exp_urcalls))
            self.assertEqual(sorted(ini_rptcalls), sorted(exp_rptcalls))

    def test_ensure_has_calls_empty(self):
        mem = chirp_common.DVMemory()
        mem.dv_urcall = 'KK7DS'
        mem.dv_rpt1call = 'KD7RFI B'
        mem.dv_rpt2call = 'KD7RFI G'
        ini_urcalls = ['', '', '', '', '']
        ini_rptcalls = ['', '', '', '', '']
        exp_urcalls = list(ini_urcalls)
        exp_rptcalls = list(ini_rptcalls)
        exp_urcalls[0] = mem.dv_urcall
        exp_rptcalls[0] = mem.dv_rpt1call
        exp_rptcalls[1] = mem.dv_rpt2call
        self._test_ensure_has_calls(mem, ini_urcalls, ini_rptcalls,
                                    exp_urcalls, exp_rptcalls)

    def test_ensure_has_calls_partial(self):
        mem = chirp_common.DVMemory()
        mem.dv_urcall = 'KK7DS'
        mem.dv_rpt1call = 'KD7RFI B'
        mem.dv_rpt2call = 'KD7RFI G'
        ini_urcalls = ['FOO', 'BAR', '', '', '']
        ini_rptcalls = ['FOO', 'BAR', '', '', '']
        exp_urcalls = list(ini_urcalls)
        exp_rptcalls = list(ini_rptcalls)
        exp_urcalls[2] = mem.dv_urcall
        exp_rptcalls[2] = mem.dv_rpt1call
        exp_rptcalls[3] = mem.dv_rpt2call
        self._test_ensure_has_calls(mem, ini_urcalls, ini_rptcalls,
                                    exp_urcalls, exp_rptcalls)

    def test_ensure_has_calls_almost_full(self):
        mem = chirp_common.DVMemory()
        mem.dv_urcall = 'KK7DS'
        mem.dv_rpt1call = 'KD7RFI B'
        mem.dv_rpt2call = 'KD7RFI G'
        ini_urcalls = ['FOO', 'BAR', 'BAZ', 'BAT', '']
        ini_rptcalls = ['FOO', 'BAR', 'BAZ', '', '']
        exp_urcalls = list(ini_urcalls)
        exp_rptcalls = list(ini_rptcalls)
        exp_urcalls[4] = mem.dv_urcall
        exp_rptcalls[3] = mem.dv_rpt1call
        exp_rptcalls[4] = mem.dv_rpt2call
        self._test_ensure_has_calls(mem, ini_urcalls, ini_rptcalls,
                                    exp_urcalls, exp_rptcalls)

    def test_ensure_has_calls_urcall_full(self):
        mem = chirp_common.DVMemory()
        mem.dv_urcall = 'KK7DS'
        mem.dv_rpt1call = 'KD7RFI B'
        mem.dv_rpt2call = 'KD7RFI G'
        ini_urcalls = ['FOO', 'BAR', 'BAZ', 'BAT', 'BOOM']
        ini_rptcalls = ['FOO', 'BAR', 'BAZ', '', '']
        exp_urcalls = list(ini_urcalls)
        exp_rptcalls = list(ini_rptcalls)
        exp_urcalls[4] = mem.dv_urcall
        exp_rptcalls[3] = mem.dv_rpt1call
        exp_rptcalls[4] = mem.dv_rpt2call
        self.assertRaises(errors.RadioError,
                          self._test_ensure_has_calls,
                          mem, ini_urcalls, ini_rptcalls,
                          exp_urcalls, exp_rptcalls)

    def test_ensure_has_calls_rptcall_full1(self):
        mem = chirp_common.DVMemory()
        mem.dv_urcall = 'KK7DS'
        mem.dv_rpt1call = 'KD7RFI B'
        mem.dv_rpt2call = 'KD7RFI G'
        ini_urcalls = ['FOO', 'BAR', 'BAZ', 'BAT', '']
        ini_rptcalls = ['FOO', 'BAR', 'BAZ', 'BAT', '']
        exp_urcalls = list(ini_urcalls)
        exp_rptcalls = list(ini_rptcalls)
        exp_urcalls[4] = mem.dv_urcall
        exp_rptcalls[3] = mem.dv_rpt1call
        exp_rptcalls[4] = mem.dv_rpt2call
        self.assertRaises(errors.RadioError,
                          self._test_ensure_has_calls,
                          mem, ini_urcalls, ini_rptcalls,
                          exp_urcalls, exp_rptcalls)

    def test_ensure_has_calls_rptcall_full2(self):
        mem = chirp_common.DVMemory()
        mem.dv_urcall = 'KK7DS'
        mem.dv_rpt1call = 'KD7RFI B'
        mem.dv_rpt2call = 'KD7RFI G'
        ini_urcalls = ['FOO', 'BAR', 'BAZ', 'BAT', '']
        ini_rptcalls = ['FOO', 'BAR', 'BAZ', 'BAT', 'BOOM']
        exp_urcalls = list(ini_urcalls)
        exp_rptcalls = list(ini_rptcalls)
        exp_urcalls[4] = mem.dv_urcall
        exp_rptcalls[3] = mem.dv_rpt1call
        exp_rptcalls[4] = mem.dv_rpt2call
        self.assertRaises(errors.RadioError,
                          self._test_ensure_has_calls,
                          mem, ini_urcalls, ini_rptcalls,
                          exp_urcalls, exp_rptcalls)


class ImportFieldTests(base.BaseTest):
    def test_import_freq(self):
        mem = chirp_common.Memory()
        mem.freq = 10
        self.assertRaises(import_logic.DestNotCompatible,
                          import_logic._import_freq,
                          FakeRadio(None), None, mem)

    def test_import_name(self):
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.name = 'foo'
        import_logic._import_name(FakeRadio(None), None, mem)
        self.assertEqual(mem.name, 'filtered-name')

    def test_import_power_same(self):
        radio = FakeRadio(None)
        same_rf = radio.get_features()
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.power = same_rf.valid_power_levels[0]
        import_logic._import_power(radio, same_rf, mem)
        self.assertEqual(mem.power, same_rf.valid_power_levels[0])

    def test_import_power_no_src(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.power = None
        import_logic._import_power(radio, src_rf, mem)
        self.assertEqual(mem.power, radio.POWER_LEVELS[1])

    def test_import_power_no_src_variable(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.power = None
        with mock.patch.object(radio, 'get_features') as mock_gf:
            mock_gf.return_value = chirp_common.RadioFeatures()
            mock_gf.return_value.has_variable_power = True
            mock_gf.return_value.valid_power_levels = [
                chirp_common.AutoNamedPowerLevel(1),
                chirp_common.AutoNamedPowerLevel(50),
                chirp_common.AutoNamedPowerLevel(1500)]
            import_logic._import_power(radio, src_rf, mem)
        self.assertEqual('50W', str(mem.power))

    def test_import_power_no_dst(self):
        radio = FakeRadio(None)
        src_rf = radio.get_features()  # Steal a copy before we stub out
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.power = src_rf.valid_power_levels[0]
        with mock.patch.object(radio, 'get_features') as mock_gf:
            mock_gf.return_value = chirp_common.RadioFeatures()
            import_logic._import_power(radio, src_rf, mem)
        self.assertEqual(mem.power, None)

    def test_import_power_closest(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        src_rf.valid_power_levels = [
            chirp_common.PowerLevel('foo', watts=7),
            chirp_common.PowerLevel('bar', watts=51),
            chirp_common.PowerLevel('baz', watts=1),
        ]
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.power = src_rf.valid_power_levels[0]
        import_logic._import_power(radio, src_rf, mem)
        self.assertEqual(mem.power, radio.POWER_LEVELS[0])

    def test_import_power_closest_watts(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        src_rf.valid_power_levels = [
            chirp_common.PowerLevel('foo', watts=20),
            chirp_common.PowerLevel('bar', watts=51),
            chirp_common.PowerLevel('baz', watts=1),
        ]
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.power = src_rf.valid_power_levels[0]
        import_logic._import_power(radio, src_rf, mem)
        self.assertEqual(mem.power, radio.POWER_LEVELS[0])

    def test_import_tone_diffA_tsql(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        src_rf.has_ctone = False
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.tmode = 'TSQL'
        mem.rtone = 100.0
        import_logic._import_tone(radio, src_rf, mem)
        self.assertEqual(mem.ctone, 100.0)

    def test_import_tone_diffB_tsql(self):
        radio = FakeRadio(None)
        radio.HAS_CTONE = False
        src_rf = chirp_common.RadioFeatures()
        src_rf.has_ctone = True
        mem = chirp_common.Memory()
        mem.tmode = 'TSQL'
        mem.ctone = 100.0
        import_logic._import_tone(radio, src_rf, mem)
        self.assertEqual(mem.rtone, 100.0)

    def test_import_dtcs_diffA_dtcs(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        src_rf.has_rx_dtcs = True
        mem = chirp_common.Memory()
        mem.tmode = 'DTCS'
        mem.rx_dtcs = 32
        import_logic._import_dtcs(radio, src_rf, mem)
        self.assertEqual(mem.dtcs, 32)

    def test_import_dtcs_diffB_dtcs(self):
        radio = FakeRadio(None)
        radio.HAS_RX_DTCS = True
        src_rf = chirp_common.RadioFeatures()
        src_rf.has_rx_dtcs = False
        mem = chirp_common.Memory()
        mem.tmode = 'DTCS'
        mem.dtcs = 32
        import_logic._import_dtcs(radio, src_rf, mem)
        self.assertEqual(mem.rx_dtcs, 32)

    def test_import_mode_valid_fm(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.mode = 'Auto'
        mem.freq = 146000000
        import_logic._import_mode(radio, None, mem)
        self.assertEqual(mem.mode, 'FM')

    def test_import_mode_valid_am(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.mode = 'Auto'
        mem.freq = 18000000
        import_logic._import_mode(radio, None, mem)
        self.assertEqual(mem.mode, 'AM')

    def test_import_mode_invalid(self):
        radio = FakeRadio(None)
        radio.MODES.remove('AM')
        mem = chirp_common.Memory()
        mem.mode = 'Auto'
        mem.freq = 1800000
        self.assertRaises(import_logic.DestNotCompatible,
                          import_logic._import_mode, radio, None, mem)

    def test_import_duplex_vhf(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.freq = 146000000
        mem.offset = 146600000
        mem.duplex = 'split'
        import_logic._import_duplex(radio, None, mem)
        self.assertEqual(mem.duplex, '+')
        self.assertEqual(mem.offset, 600000)

    def test_import_duplex_negative(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.freq = 146600000
        mem.offset = 146000000
        mem.duplex = 'split'
        import_logic._import_duplex(radio, None, mem)
        self.assertEqual(mem.duplex, '-')
        self.assertEqual(mem.offset, 600000)

    def test_import_duplex_uhf(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.freq = 431000000
        mem.offset = 441000000
        mem.duplex = 'split'
        import_logic._import_duplex(radio, None, mem)
        self.assertEqual(mem.duplex, '+')
        self.assertEqual(mem.offset, 10000000)

    def test_import_duplex_too_big_vhf(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.freq = 146000000
        mem.offset = 246000000
        mem.duplex = 'split'
        self.assertRaises(import_logic.DestNotCompatible,
                          import_logic._import_duplex, radio, None, mem)

    def test_import_duplex_off_not_supported(self):
        radio = FakeRadio(None)
        mem = chirp_common.Memory()
        mem.freq = 146000000
        mem.duplex = 'off'
        import_logic._import_duplex(radio, None, mem)
        self.assertEqual('', mem.duplex)

    @mock.patch('chirp.import_logic._import_name')
    @mock.patch('chirp.import_logic._import_power')
    @mock.patch('chirp.import_logic._import_tone')
    @mock.patch('chirp.import_logic._import_dtcs')
    @mock.patch('chirp.import_logic._import_mode')
    @mock.patch('chirp.import_logic._import_duplex')
    def _test_import_mem(self, errors,
                         mock_duplex, mock_mode, mock_dtcs, mock_tone,
                         mock_power, mock_name):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        mem = chirp_common.Memory()
        mem.freq = 146520000

        with mock.patch.object(mem, 'dupe') as mock_dupe:
            mock_dupe.return_value = mem
            with mock.patch.object(radio, 'validate_memory') as mock_val:
                mock_val.return_value = errors
                import_logic.import_mem(radio, src_rf, mem)
                mock_val.assert_called_once_with(mem)
            mock_dupe.assert_called_once_with()

        mock_duplex.assert_called_once_with(radio, src_rf, mem)

    def test_import_mem(self):
        self._test_import_mem([])

    @mock.patch.object(FakeRadio, 'get_memory')
    @mock.patch.object(FakeRadio, 'check_set_memory_immutable_policy')
    def test_import_mem_checks_immutable(self, mock_check, mock_get):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        mem = chirp_common.Memory()
        mem.freq = 146520000
        with mock.patch.object(mem, 'dupe') as mock_dupe:
            mock_dupe.return_value = mem
            import_logic.import_mem(radio, src_rf, mem)
            mock_check.assert_called_once_with(mock_get.return_value, mem)
            mock_get.assert_called_once_with(mem.number)

    def test_import_mem_with_class(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem2 = import_logic.import_mem(radio, src_rf, mem,
                                       mem_cls=chirp_common.DVMemory)
        self.assertIsInstance(mem2, chirp_common.DVMemory)

    def test_import_mem_dstar_logic(self):
        radio = FakeRadio(None)
        src_rf = chirp_common.RadioFeatures()
        mem = chirp_common.DVMemory()
        mem.freq = 146520000
        mem.mode = 'DV'

        # non-DSTAR radio exposes only FM, so DV is not compatible
        radio.MODES = ['FM']
        self.assertRaises(import_logic.DestNotCompatible,
                          import_logic.import_mem, radio, src_rf, mem)

        # non-DSTAR radio exposes DV, so it is compatible, but obviously
        # no call list manipulation needed.
        radio.MODES = ['FM', 'DV']
        with mock.patch('chirp.import_logic.ensure_has_calls') as mock_e:
            import_logic.import_mem(radio, src_rf, mem)
            mock_e.assert_not_called()

        # A real D-STAR radio gets the full treatment, depending on the
        # required call lists flag
        radio = FakeDstarRadio(None)
        with mock.patch('chirp.import_logic.ensure_has_calls') as mock_e:
            radio.REQUIRES_CALL_LISTS = False
            import_logic.import_mem(radio, src_rf, mem)
            mock_e.assert_not_called()

            radio.REQUIRES_CALL_LISTS = True
            import_logic.import_mem(radio, src_rf, mem)
            mock_e.assert_called()

    def test_import_mem_with_warnings(self):
        self._test_import_mem([chirp_common.ValidationWarning('Test')])

    def test_import_mem_with_errors(self):
        self.assertRaises(import_logic.DestNotCompatible,
                          self._test_import_mem,
                          [chirp_common.ValidationError('Test')])

    def test_import_bank(self):
        dst_mem = chirp_common.Memory()
        dst_mem.number = 1
        src_mem = chirp_common.Memory()
        src_mem.number = 2
        dst_radio = FakeRadio(None)
        src_radio = FakeRadio(None)
        dst_bm = chirp_common.BankModel(dst_radio)
        src_bm = chirp_common.BankModel(src_radio)

        dst_banks = [chirp_common.Bank(dst_bm, 0, 'A'),
                     chirp_common.Bank(dst_bm, 1, 'B'),
                     chirp_common.Bank(dst_bm, 2, 'C'),
                     ]
        src_banks = [chirp_common.Bank(src_bm, 1, '1'),
                     chirp_common.Bank(src_bm, 2, '2'),
                     chirp_common.Bank(src_bm, 3, '3'),
                     ]

        @mock.patch.object(dst_radio, 'get_mapping_models',
                           return_value=[dst_bm])
        @mock.patch.object(src_radio, 'get_mapping_models',
                           return_value=[src_bm])
        @mock.patch.object(dst_bm, 'get_mappings',
                           return_value=dst_banks)
        @mock.patch.object(src_bm, 'get_mappings',
                           return_value=src_banks)
        @mock.patch.object(src_bm, 'get_memory_mappings',
                           return_value=[src_banks[0]])
        @mock.patch.object(dst_bm, 'get_memory_mappings',
                           return_value=[dst_banks[1]])
        @mock.patch.object(dst_bm, 'remove_memory_from_mapping')
        @mock.patch.object(dst_bm, 'add_memory_to_mapping')
        def _test(mock_add, mock_remove, *rest):
            import_logic.import_bank(dst_radio, src_radio, dst_mem, src_mem)
            mock_remove.assert_called_once_with(dst_mem, dst_banks[1])
            mock_add.assert_called_once_with(dst_mem, dst_banks[0])

        _test()



from chirp import directory
from tests import load_tests

__builtins__['_'] = lambda s: s


class TestSuiteAdapter(object):
    """Adapter for pytest since it doesn't support the loadTests() protocol"""

    def __init__(self, locals):
        self.locals = locals

    def loadTestsFromTestCase(self, test_cls):
        self.locals[test_cls.__name__] = test_cls

    @staticmethod
    def addTests(tests):
        pass


directory.import_drivers()
adapter = TestSuiteAdapter(locals())
load_tests(adapter, None, None, suite=adapter)



# Copyright 2022 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import binascii
import os
import unittest
from unittest import mock

from chirp import directory
from chirp.drivers import alinco, alinco_dr735t
from chirp import memmap


class FakeAlincoSerial:
    """Behaves like a Serial with an Alinco radio connected"""

    def __init__(self, image):
        self.image = memmap.MemoryMapBytes(open(image, 'rb').read())
        self.readbuf = []
        self.ident = b''

    def zero(self):
        """Zero the mmap before an upload"""
        self.image = memmap.MemoryMapBytes(b'\x00' * len(self.image))

    def _handle_rw_dr735tn(self, data):
        if b'R' in data:
            addr_idx = data.index(b'R')-4
        elif b'W' in data:
            addr_idx = data.index(b'W')-4
        else:
            raise Exception('Unable to handle RW command %r' % data)
        addr = int(data[addr_idx: addr_idx+4], 16)
        cmd = chr(data[addr_idx+4])

        if cmd == 'R':
            bin_chunk = self.image[addr:addr+64]
            hex_chunk = binascii.hexlify(bin_chunk)
            assert (len(hex_chunk) == 128)
            self.readbuf.append(hex_chunk)
            self.readbuf.append(b'\r\n')

        if cmd == 'W':
            hex_chunk = data[addr_idx+5:addr_idx+128+5].decode()
            bin_chunk = binascii.unhexlify(hex_chunk)
            self.image[addr] = bin_chunk
            self.readbuf.append(b'OK\r\n')

    def _handle_rw(self, data):
        if b'R' in data:
            cmdindex = data.index(b'R')
        elif b'W' in data:
            cmdindex = data.index(b'W')
        else:
            raise Exception('Unable to handle RW command %r' % data)
        if cmdindex - 4 == 4:
            block_size = 16
        else:
            block_size = 64
        addr = int(data[4:cmdindex], 16)
        cmd = data[cmdindex:cmdindex + 1]
        if cmd == b'R':
            chunk = binascii.hexlify(self.image[addr:addr + block_size])
            if b'DJ175' in self.ident:
                # The DJ175 uses 16-byte blocks, but without the colon
                # separator
                self.readbuf.append(b'%s\r\n' % chunk)
            elif block_size == 64:
                self.readbuf.append(b'\r\n%s\r\n' % chunk)
            else:
                self.readbuf.append(b'\r\n%04X:%s\r\n' % (addr, chunk))
        elif cmd == b'W':
            chunk = binascii.unhexlify(data[cmdindex + 1:].rstrip())
            if len(chunk) != block_size:
                raise Exception('Expected %i, got %i' % (block_size,
                                                         len(chunk)))
            self.image[addr] = chunk
            if block_size == 64:
                self.readbuf.append(b'OK\r\n\r\n')
        else:
            raise Exception('Unable to handle command %s: %r' % (cmd, data))

    def write(self, data):
        """Write to the radio (i.e. buffer some response for read)"""
        # Serial port echo
        self.readbuf.append(data)

        ident_ack = b'OK\r\n\r\n'

        if data.startswith(b'DR'):
            self.ident = data.strip()
            self.readbuf.append(ident_ack)
        elif data.startswith(b'AL~DJ-G7'):
            self.ident = data.strip()
            self.readbuf.append(ident_ack)
        elif data.startswith(b'DJ'):
            self.ident = data.strip()
            self.readbuf.append(ident_ack)
        elif data.startswith(b'AL~F'):
            self._handle_rw(data)
        elif data.startswith(b'AL~E') and self.ident != b"DR735TN":
            self.readbuf.append(b'OK' * 10)
        elif data.startswith(b'AL~WHO'):
            # radio responds with model to WHO
            self.ident = b"DR735TN"
            # expect back "DR735TN\r\nOK\r\n"
            self.readbuf.append(self.ident+b"\r\n")
        elif data.startswith(b'AL~DR735J'):
            # this is somehow making the radio writeable...
            self.readbuf.append(b'OK\r\n')

        elif data.startswith(b'AL~RESET') and self.ident == b"DR735TN":
            self.readbuf.append(b'OK\r\n')

        elif data.startswith(b'AL~EEPEL'):
            self._handle_rw_dr735tn(data)

        else:
            raise Exception('Unable to handle %r' % data)

    def read(self, length):
        """Read from the radio (i.e. generate a radio's response)"""
        if len(self.readbuf[0]) == length:
            return self.readbuf.pop(0)
        else:
            raise Exception('Read of %i not match %i %r' % (
                length, len(self.readbuf[0]), self.readbuf[0]))


class AlincoCloneTest(unittest.TestCase):
    def _test_alinco_download(self, rclass, image):
        pipe = FakeAlincoSerial(image)
        radio = rclass(pipe)
        radio.sync_in()
        if not radio.NEEDS_COMPAT_SERIAL:
            self.assertIsInstance(radio._mmap, memmap.MemoryMapBytes)

    def _test_alinco_upload(self, rclass, image):
        pipe = FakeAlincoSerial(image)
        pipe.zero()
        radio = rclass(image)
        radio.pipe = pipe
        radio.sync_out()
        ref_image = open(image, 'rb').read()
        self.assertEqual(ref_image[0x200:rclass._memsize],
                         pipe.image.get_packed()[0x200:rclass._memsize])

    def _test_alinco(self, rclass):
        ident = directory.radio_class_id(rclass)
        image = os.path.join(os.path.dirname(__file__),
                             '..', 'images',
                             '%s.img' % ident)
        with mock.patch('time.sleep'):
            self._test_alinco_upload(rclass, image)
            self._test_alinco_download(rclass, image)

    def test_djg7(self):
        self._test_alinco(alinco.AlincoDJG7EG)

    def test_dr235(self):
        self._test_alinco(alinco.DR235Radio)

    def test_dr735(self):
        self._test_alinco(alinco_dr735t.AlincoDR735T)

    def test_dj175(self):
        # The 175 has a slightly different download string, so test it
        # specifically
        self._test_alinco(alinco.DJ175Radio)

    def test_all_alinco_identify(self):
        # Make sure all the alinco models have bytes for their _model
        # and can identify properly
        alincos = [x for x in directory.DRV_TO_RADIO.values()
                   if x.VENDOR == 'Alinco']
        for rclass in alincos:
            pipe = FakeAlincoSerial(__file__)
            radio = rclass(None)
            radio.pipe = pipe
            radio._identify()
            self.assertEqual(pipe.ident, radio._model)



import functools
import logging
import os
import unittest

from chirp import chirp_common
from chirp import bandplan_na

LOG = logging.getLogger(__name__)


class DriverTest(unittest.TestCase):
    RADIO_CLASS = None
    SUB_DEVICE = None
    TEST_IMAGE = None

    def setUp(self):
        super().setUp()
        self.patches = []

        if ('CHIRP_TEST_BITWISE_STRICT_BYTES' in os.environ and
                not self.RADIO_CLASS.NEEDS_COMPAT_SERIAL):
            self.use_patch(unittest.mock.patch(
                'chirp.bitwise.DataElement._compat_bytes',
                side_effect=self._strict_bytes))
            self.use_patch(unittest.mock.patch(
                'chirp.bitwise.string_straight_encode',
                side_effect=AssertionError(
                    'string_straight_encode not allowed in strict mode')))

        self.parent = self.RADIO_CLASS(self.TEST_IMAGE)
        self.parent_rf = self.parent.get_features()
        # If SUB_DEVICE is set to an index, then the actual radio we are
        # to test is get_sub_devices()[SUB_DEVICE]. Otherwise, it's the
        # actual class we were handed.
        if self.SUB_DEVICE is not None:
            self.radio = self.parent.get_sub_devices()[self.SUB_DEVICE]
            self.rf = self.radio.get_features()
        else:
            self.radio = self.parent
            self.rf = self.parent_rf

    def use_patch(self, patch):
        self.patches.append(patch)
        patch.start()

    def tearDown(self):
        for patch in self.patches:
            patch.stop()

    def _strict_bytes(self, bs, asbytes):
        """Enforce strict get_raw() behavior returning bytes()"""
        assert asbytes, 'asbytes must be True in strict mode'
        assert isinstance(bs, bytes), 'Type should be bytes here'
        return bs

    def get_mem(self):
        """Attempt to build a suitable memory for testing"""
        # Check to see if memory #1 has immutable fields, and if so,
        # use that as our template instead of constructing a memory ourselves
        try:
            m = self.radio.get_memory(1)
            # Pre-filter the name so it will match what we expect back
            if 'name' not in m.immutable:
                m.name = self.radio.filter_name(m.name)
            # Disable duplex in case it's set because this will cause some
            # weirdness if we much with other values, like offset.
            if 'duplex' not in m.immutable:
                m.duplex = ''
            if m.immutable:
                return m
        except Exception:
            pass

        m = chirp_common.Memory()

        # Some of the exposed bands may not be transmit-enabled, so
        # iterate them all
        attempt = 0
        for band_lo, band_hi in self.rf.valid_bands:
            m.freq = band_lo
            if chirp_common.is_airband(m.freq):
                # Air band requires special step handling.
                # First see if we can pick a 2m frequency to avoid airband
                # details.
                if chirp_common.in_range(146520000, self.rf.valid_bands):
                    m.freq = 146520000
                else:
                    # Nope, so go 25kHz up from the bottom of the band to
                    # avoid 8.33kHz issues
                    m.freq += 25000
            if self.rf.valid_tuning_steps:
                # If we have valid tuning steps, go one step above the
                # bottom of the band. Select a different tuning_step each
                # time, as some radios have various requirements for which
                # steps work in each band, mode, etc.
                steps = [x for x in self.rf.valid_tuning_steps
                         if x > 1]
                step_index = attempt % len(steps)
                m.tuning_step = steps[step_index]
                m.freq += int(m.tuning_step * 1000)
            elif m.freq + 1000000 < band_hi:
                # Otherwise just pick 1MHz above the bottom, which has been
                # our test basis for a long time, unless that extends past
                # the end of the band.
                m.freq += 1000000

            if m.freq < 30000000 and "AM" in self.rf.valid_modes:
                m.mode = "AM"
            else:
                try:
                    m.mode = self.rf.valid_modes[0]
                except IndexError:
                    pass

            for i in range(*self.rf.memory_bounds)[:10]:
                m.number = i
                msgs = self.radio.validate_memory(m)
                warnings, errors = chirp_common.split_validation_msgs(msgs)
                if warnings and not errors:
                    # If we got some warnings and no errors, then we know the
                    # memory is almost good enough. Set it and pull it back
                    # to let the radio squash whatever was "almost correct"
                    # and then use that.
                    self.radio.set_memory(m)
                    m = self.radio.get_memory(m.number)
                    try:
                        del m.extra
                    except AttributeError:
                        pass
                    return m
                # If we got no warnings, or we have only one band, then
                # no errors means we found our candidate.
                elif not errors:
                    return m

                attempt += 1

        self.fail("No mutable memory locations found - unable to run this "
                  "test because I don't have a memory to test with")

    def assertEqualMem(self, a, b, ignore=None):
        if a.tmode == "Cross":
            tx_mode, rx_mode = a.cross_mode.split("->")
        if ignore is None:
            ignore = []

        if a.duplex == b.duplex == 'off':
            # If we're asking for duplex=off, we should not obsess over the
            # driver keeping track of our offset, as it may use the offset
            # field to manage the TX behavior
            ignore.append('offset')

        a_vals = {}
        b_vals = {}

        for k, v in list(a.__dict__.items()):
            if ignore and k in ignore:
                continue
            if k == "power":
                continue  # FIXME
            elif k == "extra":
                continue
            elif k == "immutable":
                continue
            elif k == "name":
                if not self.rf.has_name:
                    continue  # Don't complain about name, if not supported
                else:
                    # Name mismatch fair if filter_name() is right
                    v = self.radio.filter_name(v).rstrip()
            elif k == "tuning_step" and not self.rf.has_tuning_step:
                continue
            elif k == "rtone" and not (
                a.tmode == "Tone" or
                (a.tmode == "TSQL" and not self.rf.has_ctone) or
                (a.tmode == "Cross" and tx_mode == "Tone") or
                (a.tmode == "Cross" and rx_mode == "Tone" and
                 not self.rf.has_ctone)
            ):
                continue
            elif k == "ctone" and (not self.rf.has_ctone or
                                   not (a.tmode == "TSQL" or
                                        (a.tmode == "Cross" and
                                         rx_mode == "Tone"))):
                continue
            elif k == "dtcs" and (a.tmode != 'DTCS' or
                                  (a.tmode == 'Cross' and tx_mode != 'DTCS')):
                # If we are not in a tmode where a transmit DTCS code is
                # required, we do not care if the code is persisted.
                continue
            elif k == "rx_dtcs" and (not self.rf.has_rx_dtcs or
                                     not (a.tmode == "Cross" and
                                          rx_mode == "DTCS")):
                # If we are not in a tmode where a receive DTCS code is
                # required, we do not care if the code is persisted.
                continue
            elif k == "offset" and not a.duplex:
                continue
            elif k == "cross_mode" and a.tmode != "Cross":
                continue

            if (a.freq in bandplan_na.ALL_GMRS_FREQS and
                    k in a.immutable or k in b.immutable):
                # If the radio returned a field in immutable, it probably
                # means that it's a mandatory setting (i.e. power or duplex
                # in GMRS)
                continue

            a_vals[k] = v
            b_vals[k] = b.__dict__[k]

        self.assertEqual(a_vals, b_vals,
                         'Memories have unexpected differences')

        # Consider mem.extra as matching if the structure remains the same.
        # Since we don't know anything about model-specific things here we
        # can't really assert any more than that, but we can ensure that the
        # structure doesn't change due to the contents of the rest of the
        # memory.
        if a.extra and b.extra:
            self.assertEqual([x.get_name() for x in a.extra],
                             [x.get_name() for x in b.extra],
                             'Memories have different mem.extra keys')


def requires_feature(flag, equal=None):
    def inner(fn):
        @functools.wraps(fn)
        def wraps(self, *a, **k):
            val = getattr(self.rf, flag)
            if (val == equal if equal is not None else val):
                fn(self, *a, **k)
            else:
                self.skipTest('Feature %s not supported' % flag)
        return wraps
    return inner



from unittest import mock

from tests import base


class TestCaseFeatures(base.DriverTest):
    def test_prompts(self):
        self.use_patch(mock.patch('builtins._', create=True,
                                  side_effect=lambda s: s))
        prompts = self.radio.get_prompts()
        for p in ('info', 'experimental', 'pre_download', 'pre_upload'):
            prompt = getattr(prompts, p)
            if prompt is not None:
                self.assertIsInstance(prompt, str)



import os

from chirp import chirp_common
from chirp.drivers import generic_csv
from chirp import import_logic
from tests import base


class TestCaseCopyAll(base.DriverTest):
    "Copy Memories From CSV"

    def setUp(self):
        super().setUp()
        csvfile = os.path.join(os.path.dirname(self.TEST_IMAGE),
                               'Generic_CSV.csv')
        self.src_radio = generic_csv.CSVRadio(csvfile)

    def test_copy(self):
        src_rf = self.src_radio.get_features()
        bounds = src_rf.memory_bounds

        for dst_number in range(self.rf.memory_bounds[0],
                                min(self.rf.memory_bounds[0] + 10,
                                    self.rf.memory_bounds[1])):
            cur_mem = self.radio.get_memory(dst_number)
            if not cur_mem.empty and 'freq' in cur_mem.immutable:
                # Keep looking
                continue
            else:
                break
        else:
            self.skipTest('No channels with mutable freq found to use')

        for number in range(bounds[0], bounds[1]):
            src_mem = self.src_radio.get_memory(number)
            if src_mem.empty:
                continue

            try:
                dst_mem = import_logic.import_mem(self.radio,
                                                  src_rf, src_mem,
                                                  overrides={
                                                      "number": dst_number})
                import_logic.import_bank(self.radio,
                                         self.src_radio,
                                         dst_mem,
                                         src_mem)
            except import_logic.DestNotCompatible:
                continue

            warn, err = chirp_common.split_validation_msgs(
                self.radio.validate_memory(dst_mem))
            self.radio.set_memory(dst_mem)

            if warn:
                # If the radio warned about something, we can assume it's
                # about duplex (i.e. tx inhibit) or mode (i.e. AM only on
                # airband)
                ignore = ['duplex', 'mode']
            else:
                ignore = None
            ret_mem = self.radio.get_memory(dst_number)
            self.assertEqualMem(dst_mem, ret_mem, ignore=ignore)
